
import asyncio
import re
import json
from typing import Dict, List, Optional, cast

import aiohttp
from tenacity import (
    retry,
    wait_exponential,
    stop_after_delay,
    retry_if_exception_type,
)


from .api import WebApi
from .models import LiveStatus, RoomInfo, UserInfo
from .typing import QualityNumber, StreamFormat, ResponseData
from .exceptions import (
    LiveRoomHidden, LiveRoomLocked, LiveRoomEncrypted, NoStreamUrlAvailable
)


__all__ = 'Live',


_INFO_PATTERN = re.compile(
    rb'<script>\s*window\.__NEPTUNE_IS_MY_WAIFU__\s*=\s*(\{.*?\})\s*</script>'
)
_LIVE_STATUS_PATTERN = re.compile(rb'"live_status"\s*:\s*(\d)')


class Live:
    def __init__(
        self, room_id: int, user_agent: str = '', cookie: str = ''
    ) -> None:
        self._room_id = room_id
        self._user_agent = user_agent
        self._cookie = cookie
        self._html_page_url = f'https://live.bilibili.com/{room_id}'

        self._room_info: RoomInfo
        self._user_info: UserInfo

    @property
    def user_agent(self) -> str:
        return self._user_agent

    @user_agent.setter
    def user_agent(self, value: str) -> None:
        self._user_agent = value

    @property
    def cookie(self) -> str:
        return self._cookie

    @cookie.setter
    def cookie(self, value: str) -> None:
        self._cookie = value

    @property
    def headers(self) -> Dict[str, str]:
        return {
            'Referer': 'https://live.bilibili.com/',
            'Connection': 'Keep-Alive',
            'User-Agent': self._user_agent,
            'cookie': self._cookie,
        }

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    @property
    def room_id(self) -> int:
        return self._room_id

    @property
    def room_info(self) -> RoomInfo:
        return self._room_info

    @property
    def user_info(self) -> UserInfo:
        return self._user_info

    async def init(self) -> None:
        self._session = aiohttp.ClientSession(
            headers=self.headers,
            raise_for_status=True,
            trust_env=True,
        )
        self._api = WebApi(self._session)

        self._room_info = await self.get_room_info()
        self._user_info = await self.get_user_info(self._room_info.uid)

    async def deinit(self) -> None:
        await self._session.close()

    async def get_live_status(self) -> LiveStatus:
        try:
            # frequent requests will be intercepted by the server's firewall!
            live_status = await self._get_live_status_via_api()
        except Exception:
            # more cpu consumption
            live_status = await self._get_live_status_via_html_page()

        return LiveStatus(live_status)

    def is_living(self) -> bool:
        return self._room_info.live_status == LiveStatus.LIVE

    async def update_info(self) -> None:
        await asyncio.wait([self.update_user_info(), self.update_room_info()])

    @retry(
        reraise=True,
        retry=retry_if_exception_type((
            asyncio.TimeoutError, aiohttp.ClientError,
        )),
        wait=wait_exponential(max=10),
        stop=stop_after_delay(60),
    )
    async def update_user_info(self) -> None:
        self._user_info = await self.get_user_info(self._room_info.uid)

    @retry(
        reraise=True,
        retry=retry_if_exception_type((
            asyncio.TimeoutError, aiohttp.ClientError,
        )),
        wait=wait_exponential(max=10),
        stop=stop_after_delay(60),
    )
    async def update_room_info(self) -> None:
        self._room_info = await self.get_room_info()

    async def get_room_info(self) -> RoomInfo:
        try:
            # frequent requests will be intercepted by the server's firewall!
            room_info_data = await self._get_room_info_via_api()
        except Exception:
            # more cpu consumption
            room_info_data = await self._get_room_info_via_html_page()

        return RoomInfo.from_data(room_info_data)

    async def get_user_info(self, uid: int) -> UserInfo:
        user_info_data = await self._api.get_user_info(uid)
        return UserInfo.from_data(user_info_data)

    async def get_server_timestamp(self) -> int:
        # the timestamp on the server at the moment in seconds
        return await self._api.get_timestamp()

    async def get_live_stream_url(
        self, qn: QualityNumber = 10000, format: StreamFormat = 'flv'
    ) -> Optional[str]:
        try:
            data = await self._api.get_room_play_info(self._room_id, qn)
        except Exception:
            # fallback to the html page global info
            data = await self._get_room_play_info_via_html_page()
            self._check_room_play_info(data)
            stream = data['playurl_info']['playurl']['stream']
            if stream[0]['format'][0]['codec'][0]['current_qn'] != qn:
                raise
        else:
            self._check_room_play_info(data)

        streams = list(filter(
            lambda s: s['format'][0]['format_name'] == format,
            data['playurl_info']['playurl']['stream']
        ))
        codec = streams[0]['format'][0]['codec'][0]
        accept_qn = cast(List[QualityNumber], codec['accept_qn'])

        if qn not in accept_qn:
            return None

        assert codec['current_qn'] == qn
        url_info = codec['url_info'][0]

        return url_info['host'] + codec['base_url'] + url_info['extra']

    def _check_room_play_info(self, data: ResponseData) -> None:
        if data['is_hidden']:
            raise LiveRoomHidden()
        if data['is_locked']:
            raise LiveRoomLocked()
        if data['encrypted'] and not data['pwd_verified']:
            raise LiveRoomEncrypted()
        try:
            data['playurl_info']['playurl']['stream'][0]
        except Exception:
            raise NoStreamUrlAvailable()

    async def _get_live_status_via_api(self) -> int:
        room_info_data = await self._get_room_info_via_api()
        return int(room_info_data['live_status'])

    async def _get_room_info_via_api(self) -> ResponseData:
        try:
            room_info_data = await self._api.get_info(self._room_id)
        except Exception:
            info_data = await self._api.get_info_by_room(self._room_id)
            room_info_data = info_data['room_info']

        return room_info_data

    async def _get_live_status_via_html_page(self) -> int:
        async with self._session.get(self._html_page_url) as response:
            data = await response.read()

        m = _LIVE_STATUS_PATTERN.search(data)
        assert m is not None, data

        return int(m.group(1))

    async def _get_room_info_via_html_page(self) -> ResponseData:
        info_res = await self._get_room_info_res_via_html_page()
        return info_res['room_info']

    async def _get_room_play_info_via_html_page(self) -> ResponseData:
        return await self._get_room_init_res_via_html_page()

    async def _get_room_info_res_via_html_page(self) -> ResponseData:
        info = await self._get_info_via_html_page()
        if info['roomInfoRes']['code'] != 0:
            raise ValueError(f"Invaild roomInfoRes: {info['roomInfoRes']}")
        return info['roomInfoRes']['data']

    async def _get_room_init_res_via_html_page(self) -> ResponseData:
        info = await self._get_info_via_html_page()
        if info['roomInitRes']['code'] != 0:
            raise ValueError(f"Invaild roomInitRes: {info['roomInitRes']}")
        return info['roomInitRes']['data']

    async def _get_info_via_html_page(self) -> ResponseData:
        async with self._session.get(self._html_page_url) as response:
            data = await response.read()

        m = _INFO_PATTERN.search(data)
        assert m is not None, data

        string = m.group(1).decode(encoding='utf8')
        return json.loads(string)



from typing import AbstractSet, Literal


KeyOfSettings = Literal[
    'version',
    'tasks',
    'output',
    'logging',
    'header',
    'danmaku',
    'recorder',
    'postprocessing',
    'space',
    'email_notification',
    'serverchan_notification',
    'pushplus_notification',
    'webhooks',
]

KeySetOfSettings = AbstractSet[KeyOfSettings]



import asyncio
import logging
from pathlib import PurePath
from typing import (
    Any, Awaitable, Dict, Iterable, Iterator, List, Optional, Sequence, Set,
    Tuple, Union,
)

from rx.core.typing import Scheduler
from rx.scheduler.threadpoolscheduler import ThreadPoolScheduler

from .models import ProcessStatus, DeleteStrategy
from .typing import Progress
from .remuxer import remux_video, RemuxProgress, RemuxResult
from .helpers import discard_file, discard_files, get_extra_metadata
from .ffmpeg_metadata import make_metadata_file
from ..event.event_emitter import EventListener, EventEmitter
from ..bili.live import Live
from ..core import Recorder, RecorderEventListener
from ..exception import exception_callback
from ..utils.mixins import SwitchableMixin, AsyncCooperationMix
from ..path import danmaku_path, extra_metadata_path
from ..flv.metadata_injector import inject_metadata, InjectProgress
from ..flv.helpers import is_valid
from ..logging.room_id import aio_task_with_room_id


__all__ = (
    'Postprocessor',
    'PostprocessorEventListener',
    'ProcessStatus',
    'DeleteStrategy',
)


logger = logging.getLogger(__name__)


class PostprocessorEventListener(EventListener):
    async def on_file_completed(self, room_id: int, path: str) -> None:
        ...


class Postprocessor(
    EventEmitter[PostprocessorEventListener],
    RecorderEventListener,
    SwitchableMixin,
    AsyncCooperationMix,
):
    def __init__(
        self,
        live: Live,
        recorder: Recorder,
        *,
        remux_to_mp4: bool = False,
        delete_source: DeleteStrategy = DeleteStrategy.AUTO,
    ) -> None:
        super().__init__()

        self._live = live
        self._recorder = recorder

        self.remux_to_mp4 = remux_to_mp4
        self.delete_source = delete_source

        self._status = ProcessStatus.WAITING
        self._postprocessing_path: Optional[str] = None
        self._postprocessing_progress: Optional[Progress] = None
        self._files: List[str] = []

    @property
    def status(self) -> ProcessStatus:
        return self._status

    @property
    def postprocessing_path(self) -> Optional[str]:
        return self._postprocessing_path

    @property
    def postprocessing_progress(self) -> Optional[Progress]:
        return self._postprocessing_progress

    def get_final_files(self) -> Iterator[str]:
        yield from iter(self._files)

    async def wait(self) -> None:
        await self._wait_processing_task()

    async def on_recording_finished(self, recorder: Recorder) -> None:
        self._create_processing_task(recorder)

    async def on_recording_cancelled(self, recorder: Recorder) -> None:
        self._create_processing_task(recorder)

    def _do_enable(self) -> None:
        self._recorder.add_listener(self)
        logger.debug('Enabled postprocessor')

    def _do_disable(self) -> None:
        self._recorder.remove_listener(self)
        logger.debug('Disabled postprocessor')

    def _create_processing_task(self, recorder: Recorder) -> None:
        raw_videos = list(recorder.get_video_files())
        raw_danmaku = list(recorder.get_danmaku_files())
        self._processing_task = asyncio.create_task(
            self._process(raw_videos, raw_danmaku),
        )
        self._processing_task.add_done_callback(exception_callback)

    async def _wait_processing_task(self) -> None:
        if hasattr(self, '_processing_task'):
            await self._processing_task

    @aio_task_with_room_id
    async def _process(
        self, raw_videos: Sequence[str], raw_danmaku: Sequence[str]
    ) -> None:
        raw_videos, raw_danmaku = await self._discard_invaild_files(
            raw_videos, raw_danmaku
        )

        self._files.clear()

        if self.remux_to_mp4:
            self._status = ProcessStatus.REMUXING
            final_videos = await self._remux_videos(raw_videos)
        else:
            self._status = ProcessStatus.INJECTING
            await self._inject_extra_metadata(raw_videos)
            final_videos = raw_danmaku

        self._files.extend(final_videos)
        self._files.extend(raw_danmaku)

        self._status = ProcessStatus.WAITING
        self._postprocessing_path = None
        self._postprocessing_progress = None

    async def _discard_invaild_files(
        self, raw_videos: Sequence[str], raw_danmaku: Sequence[str]
    ) -> Tuple[List[str], List[str]]:
        loop = asyncio.get_running_loop()
        valid_video_set: Set[str] = await loop.run_in_executor(
            None, set, filter(is_valid, raw_videos)
        )
        invalid_video_set = set(raw_videos) - valid_video_set

        if invalid_video_set:
            logger.info('Discarding invalid files ...')
            await discard_files(invalid_video_set)
            await discard_files(map(danmaku_path, invalid_video_set))

        return list(valid_video_set), list(map(danmaku_path, valid_video_set))

    async def _inject_extra_metadata(self, video_paths: Iterable[str]) -> None:
        scheduler = ThreadPoolScheduler()
        for path in video_paths:
            metadata = await get_extra_metadata(path)
            await self._inject_metadata(path, metadata, scheduler)
            await discard_file(extra_metadata_path(path), 'DEBUG')
            await self._emit('file_completed', self._live.room_id, path)

    async def _remux_videos(self, video_paths: Iterable[str]) -> List[str]:
        results = []
        scheduler = ThreadPoolScheduler()

        for in_path in video_paths:
            out_path = str(PurePath(in_path).with_suffix('.mp4'))
            logger.info(f"Remuxing '{in_path}' to '{out_path}' ...")

            metadata_path = await make_metadata_file(in_path)
            remux_result = await self._remux_video(
                in_path, out_path, metadata_path, scheduler
            )

            if remux_result.is_successful():
                logger.info(f"Successfully remux '{in_path}' to '{out_path}'")
                result_path = out_path
            elif remux_result.is_warned():
                logger.warning('Remuxing done, but ran into problems.')
                result_path = out_path
            elif remux_result.is_failed:
                logger.error(f"Failed to remux '{in_path}' to '{out_path}'")
                result_path = in_path
            else:
                pass

            logger.debug(f'ffmpeg output:\n{remux_result.output}')

            if self._should_delete_source_files(remux_result):
                await discard_file(in_path)
                await discard_files(
                    [metadata_path, extra_metadata_path(in_path)], 'DEBUG'
                )

            results.append(result_path)
            await self._emit('file_completed', self._live.room_id, result_path)

        return results

    def _inject_metadata(
        self,
        path: str,
        metadata: Dict[str, Any],
        scheduler: Scheduler,
    ) -> Awaitable[None]:
        future: asyncio.Future[None] = asyncio.Future()
        self._postprocessing_path = path

        def on_next(value: InjectProgress) -> None:
            self._postprocessing_progress = value

        inject_metadata(
            path,
            metadata,
            report_progress=True,
            room_id=self._live.room_id,
        ).subscribe(
            on_next,
            lambda e: future.set_exception(e),
            lambda: future.set_result(None),
            scheduler=scheduler,
        )

        return future

    def _remux_video(
        self,
        in_path: str,
        out_path: str,
        metadata_path: str,
        scheduler: Scheduler,
    ) -> Awaitable[RemuxResult]:
        future: asyncio.Future[RemuxResult] = asyncio.Future()
        self._postprocessing_path = in_path

        def on_next(value: Union[RemuxProgress, RemuxResult]) -> None:
            if isinstance(value, RemuxProgress):
                self._postprocessing_progress = value
            elif isinstance(value, RemuxResult):
                future.set_result(value)

        remux_video(
            in_path,
            out_path,
            metadata_path,
            report_progress=True,
        ).subscribe(
            on_next,
            lambda e: future.set_exception(e),
            scheduler=scheduler,
        )

        return future

    def _should_delete_source_files(
        self, remux_result: RemuxResult
    ) -> bool:
        if self.delete_source == DeleteStrategy.AUTO:
            if not remux_result.is_failed():
                return True
        elif self.delete_source == DeleteStrategy.NEVER:
            return False

        return False



import os
import logging
import asyncio
from typing import Iterator, List, Optional

import attr
import psutil

from . import __prog__, __version__
from .disk_space import SpaceMonitor, SpaceReclaimer
from .bili.helpers import ensure_room_id
from .task import (
    RecordTaskManager,
    TaskParam,
    TaskData,
    FileDetail,
)
from .exception import ExistsError, ExceptionHandler
from .event.event_submitters import SpaceEventSubmitter
from .setting import (
    SettingsManager,
    Settings,
    SettingsIn,
    SettingsOut,
    KeySetOfSettings,
    TaskOptions,
)
from .notification import (
    EmailNotifier,
    ServerchanNotifier,
    PushplusNotifier,
)
from .webhook import WebHookEmitter


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class AppInfo:
    name: str
    version: str
    pid: int
    ppid: int
    create_time: float
    cwd: str
    exe: str
    cmdline: List[str]


@attr.s(auto_attribs=True, slots=True, frozen=True)
class AppStatus:
    cpu_percent: float
    memory_percent: float
    num_threads: int


class Application:
    def __init__(self, settings: Settings) -> None:
        self._out_dir = settings.output.out_dir
        self._settings_manager = SettingsManager(self, settings)
        self._task_manager = RecordTaskManager(self._settings_manager)

    @property
    def info(self) -> AppInfo:
        p = psutil.Process(os.getpid())
        with p.oneshot():
            return AppInfo(
                name=__prog__,
                version=__version__,
                pid=p.pid,
                ppid=p.ppid(),
                create_time=p.create_time(),
                cwd=p.cwd(),
                exe=p.exe(),
                cmdline=p.cmdline(),
            )

    @property
    def status(self) -> AppStatus:
        p = psutil.Process(os.getpid())
        with p.oneshot():
            return AppStatus(
                cpu_percent=p.cpu_percent(),
                memory_percent=p.memory_percent(),
                num_threads=p.num_threads(),
            )

    def run(self) -> None:
        asyncio.run(self._run())

    async def _run(self) -> None:
        self._loop = asyncio.get_running_loop()

        await self.launch()
        try:
            self._interrupt_event = asyncio.Event()
            await self._interrupt_event.wait()
        finally:
            await self.exit()

    async def launch(self) -> None:
        self._setup()
        await self._task_manager.load_all_tasks()
        logger.info('Launched Application')

    async def exit(self) -> None:
        await self._exit()
        logger.info('Exited Application')

    async def abort(self) -> None:
        await self._exit(force=True)
        logger.info('Aborted Application')

    async def _exit(self, force: bool = False) -> None:
        await self._task_manager.stop_all_tasks(force=force)
        await self._task_manager.destroy_all_tasks()
        self._destroy()

    async def restart(self) -> None:
        logger.info('Restarting Application...')
        await self.exit()
        await self.launch()

    def has_task(self, room_id: int) -> bool:
        return self._task_manager.has_task(room_id)

    async def add_task(self, room_id: int) -> int:
        room_id = await ensure_room_id(room_id)

        if self._settings_manager.has_task_settings(room_id):
            raise ExistsError(
                f"a task for the room {room_id} is already existed"
            )

        settings = await self._settings_manager.add_task_settings(room_id)
        await self._task_manager.add_task(settings)

        return room_id

    async def remove_task(self, room_id: int) -> None:
        await self._task_manager.remove_task(room_id)
        await self._settings_manager.remove_task_settings(room_id)

    async def remove_all_tasks(self) -> None:
        await self._task_manager.remove_all_tasks()
        await self._settings_manager.remove_all_task_settings()

    async def start_task(self, room_id: int) -> None:
        await self._task_manager.start_task(room_id)
        await self._settings_manager.mark_task_enabled(room_id)

    async def stop_task(self, room_id: int, force: bool = False) -> None:
        await self._task_manager.stop_task(room_id, force)
        await self._settings_manager.mark_task_disabled(room_id)

    async def start_all_tasks(self) -> None:
        await self._task_manager.start_all_tasks()
        await self._settings_manager.mark_all_tasks_enabled()

    async def stop_all_tasks(self, force: bool = False) -> None:
        await self._task_manager.stop_all_tasks(force)
        await self._settings_manager.mark_all_tasks_disabled()

    async def enable_task_recorder(self, room_id: int) -> None:
        await self._task_manager.enable_task_recorder(room_id)
        await self._settings_manager.mark_task_recorder_enabled(room_id)

    async def disable_task_recorder(
        self, room_id: int, force: bool = False
    ) -> None:
        await self._task_manager.disable_task_recorder(room_id, force)
        await self._settings_manager.mark_task_recorder_disabled(room_id)

    async def enable_all_task_recorders(self) -> None:
        await self._task_manager.enable_all_task_recorders()
        await self._settings_manager.mark_all_task_recorders_enabled()

    async def disable_all_task_recorders(self, force: bool = False) -> None:
        await self._task_manager.disable_all_task_recorders(force)
        await self._settings_manager.mark_all_task_recorders_disabled()

    def get_task_data(self, room_id: int) -> TaskData:
        return self._task_manager.get_task_data(room_id)

    def get_all_task_data(self) -> Iterator[TaskData]:
        yield from self._task_manager.get_all_task_data()

    def get_task_param(self, room_id: int) -> TaskParam:
        return self._task_manager.get_task_param(room_id)

    def get_task_file_details(self, room_id: int) -> Iterator[FileDetail]:
        yield from self._task_manager.get_task_file_details(room_id)

    async def update_task_info(self, room_id: int) -> None:
        await self._task_manager.update_task_info(room_id)

    async def update_all_task_infos(self) -> None:
        await self._task_manager.update_all_task_infos()

    def get_settings(
        self,
        include: Optional[KeySetOfSettings] = None,
        exclude: Optional[KeySetOfSettings] = None,
    ) -> SettingsOut:
        return self._settings_manager.get_settings(include, exclude)

    async def change_settings(self, settings: SettingsIn) -> SettingsOut:
        return await self._settings_manager.change_settings(settings)

    def get_task_options(self, room_id: int) -> TaskOptions:
        return self._settings_manager.get_task_options(room_id)

    async def change_task_options(
        self, room_id: int, options: TaskOptions
    ) -> TaskOptions:
        return await self._settings_manager.change_task_options(
            room_id, options
        )

    def _setup(self) -> None:
        self._setup_logger()
        self._setup_exception_handler()
        self._setup_space_monitor()
        self._setup_space_event_submitter()
        self._setup_space_reclaimer()
        self._setup_notifiers()
        self._setup_webhooks()

    def _setup_logger(self) -> None:
        self._settings_manager.apply_logging_settings()

    def _setup_exception_handler(self) -> None:
        self._exception_handler = ExceptionHandler()
        self._exception_handler.enable()

    def _setup_space_monitor(self) -> None:
        self._space_monitor = SpaceMonitor(self._out_dir)
        self._settings_manager.apply_space_monitor_settings()
        self._space_monitor.enable()

    def _setup_space_event_submitter(self) -> None:
        self._space_event_submitter = SpaceEventSubmitter(self._space_monitor)

    def _setup_space_reclaimer(self) -> None:
        self._space_reclaimer = SpaceReclaimer(
            self._space_monitor, self._out_dir,
        )
        self._settings_manager.apply_space_reclaimer_settings()
        self._space_reclaimer.enable()

    def _setup_notifiers(self) -> None:
        self._email_notifier = EmailNotifier()
        self._serverchan_notifier = ServerchanNotifier()
        self._pushplus_notifier = PushplusNotifier()
        self._settings_manager.apply_email_notification_settings()
        self._settings_manager.apply_serverchan_notification_settings()
        self._settings_manager.apply_pushplus_notification_settings()

    def _setup_webhooks(self) -> None:
        self._webhook_emitter = WebHookEmitter()
        self._settings_manager.apply_webhooks_settings()
        self._webhook_emitter.enable()

    def _destroy(self) -> None:
        self._destroy_space_reclaimer()
        self._destroy_space_event_submitter()
        self._destroy_space_monitor()
        self._destroy_notifiers()
        self._destroy_webhooks()
        self._destroy_exception_handler()

    def _destroy_space_monitor(self) -> None:
        self._space_monitor.disable()
        del self._space_monitor

    def _destroy_space_event_submitter(self) -> None:
        del self._space_event_submitter

    def _destroy_space_reclaimer(self) -> None:
        self._space_reclaimer.disable()
        del self._space_reclaimer

    def _destroy_notifiers(self) -> None:
        self._email_notifier.disable()
        self._serverchan_notifier.disable()
        self._pushplus_notifier.disable()
        del self._email_notifier
        del self._serverchan_notifier
        del self._pushplus_notifier

    def _destroy_webhooks(self) -> None:
        self._webhook_emitter.disable()
        del self._webhook_emitter

    def _destroy_exception_handler(self) -> None:
        self._exception_handler.disable()
        del self._exception_handler



from .event_center import EventCenter
from .event_emitter import EventEmitter, EventListener
from .models import (
    BaseEvent,
    BaseEventData,
    LiveBeganEvent,
    LiveBeganEventData,
    LiveEndedEvent,
    LiveEndedEventData,
    RoomChangeEvent,
    RoomChangeEventData,
    SpaceNoEnoughEvent,
    SpaceNoEnoughEventData,
    FileCompletedEvent,
    FileCompletedEventData,
    Error,
    ErrorData,
)


__all__ = (
    'EventCenter',
    'EventEmitter', 
    'EventListener',

    'BaseEvent',
    'BaseEventData',
    'LiveBeganEvent',
    'LiveBeganEventData',
    'LiveEndedEvent',
    'LiveEndedEventData',
    'RoomChangeEvent',
    'RoomChangeEventData',
    'SpaceNoEnoughEvent',
    'SpaceNoEnoughEventData',
    'FileCompletedEvent',
    'FileCompletedEventData',
    'Error',
    'ErrorData',
)



import html
import asyncio
import logging
from contextlib import suppress
from typing import Iterator, List


from .. import __version__, __prog__, __github__
from .danmaku_receiver import DanmakuReceiver, DanmuMsg
from .stream_recorder import StreamRecorder, StreamRecorderEventListener
from .statistics import StatisticsCalculator
from ..bili.live import Live
from ..exception import exception_callback
from ..path import danmaku_path
from ..danmaku.models import Metadata, Danmu
from ..danmaku.io import DanmakuWriter
from ..utils.mixins import SwitchableMixin
from ..logging.room_id import aio_task_with_room_id


__all__ = 'DanmakuDumper',


logger = logging.getLogger(__name__)


class DanmakuDumper(StreamRecorderEventListener, SwitchableMixin):
    def __init__(
        self,
        live: Live,
        stream_recorder: StreamRecorder,
        danmaku_receiver: DanmakuReceiver,
        *,
        danmu_uname: bool = False,
    ) -> None:
        super().__init__()

        self._live = live
        self._stream_recorder = stream_recorder
        self._receiver = danmaku_receiver

        self.danmu_uname = danmu_uname

        self._files: List[str] = []
        self._calculator = StatisticsCalculator(interval=60)

    @property
    def danmu_count(self) -> int:
        return self._calculator.count

    @property
    def danmu_rate(self) -> float:
        return self._calculator.rate

    @property
    def elapsed(self) -> float:
        return self._calculator.elapsed

    def _do_enable(self) -> None:
        self._stream_recorder.add_listener(self)
        logger.debug('Enabled danmaku dumper')

    def _do_disable(self) -> None:
        self._stream_recorder.remove_listener(self)
        logger.debug('Disabled danmaku dumper')

    def set_live_start_time(self, time: int) -> None:
        self._live_start_time = time

    def has_file(self) -> bool:
        return bool(self._files)

    def get_files(self) -> Iterator[str]:
        for file in self._files:
            yield file

    def clear_files(self) -> None:
        self._files.clear()

    async def on_video_file_created(
        self, video_path: str, record_start_time: int
    ) -> None:
        self._path = danmaku_path(video_path)
        self._record_start_time = record_start_time
        self._files.append(self._path)
        self._start_dumping()

    async def on_video_file_completed(self, video_path: str) -> None:
        await self._stop_dumping()

    def _start_dumping(self) -> None:
        self._create_dump_task()

    async def _stop_dumping(self) -> None:
        await self._cancel_dump_task()

    def _create_dump_task(self) -> None:
        self._dump_task = asyncio.create_task(self._dump())
        self._dump_task.add_done_callback(exception_callback)

    async def _cancel_dump_task(self) -> None:
        self._dump_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._dump_task

    @aio_task_with_room_id
    async def _dump(self) -> None:
        logger.debug('Started dumping danmaku')
        self._calculator.reset()

        try:
            async with DanmakuWriter(self._path) as writer:
                logger.info(f"Danmaku file created: '{self._path}'")
                await writer.write_metadata(self._make_metadata())

                while True:
                    msg = await self._receiver.get_message()
                    await writer.write_danmu(self._make_danmu(msg))
                    self._calculator.submit(1)
        finally:
            logger.info(f"Danmaku file completed: '{self._path}'")
            logger.debug('Stopped dumping danmaku')
            self._calculator.freeze()

    def _make_metadata(self) -> Metadata:
        return Metadata(
            user_name=self._live.user_info.name,
            room_id=self._live.room_info.room_id,
            room_title=self._live.room_info.title,
            area=self._live.room_info.area_name,
            parent_area=self._live.room_info.parent_area_name,
            live_start_time=self._live.room_info.live_start_time,
            record_start_time=self._record_start_time,
            recorder=f'{__prog__} v{__version__} {__github__}',
        )

    def _make_danmu(self, msg: DanmuMsg) -> Danmu:
        stime = max((msg.date - self._record_start_time * 1000), 0) / 1000

        if self.danmu_uname:
            text = f'{msg.uname}: {msg.text}'
        else:
            text = msg.text
        text = html.escape(text)

        return Danmu(
            stime=stime,
            mode=msg.mode,
            size=msg.size,
            color=msg.color,
            date=msg.date,
            pool=msg.pool,
            uid=msg.uid,
            dmid=msg.dmid,
            text=text,
        )



from .models import (
    DEFAULT_SETTINGS_PATH,

    EnvSettings,
    Settings,
    SettingsIn,
    SettingsOut,

    HeaderOptions,
    HeaderSettings,
    DanmakuOptions,
    DanmakuSettings,
    RecorderOptions,
    RecorderSettings,
    PostprocessingSettings,
    PostprocessingOptions,

    TaskOptions,
    TaskSettings,
    OutputSettings,
    LoggingSettings,
    SpaceSettings,
    EmailSettings,
    ServerchanSettings,
    PushplusSettings,
    NotifierSettings,
    NotificationSettings,
    EmailNotificationSettings,
    ServerchanNotificationSettings,
    PushplusNotificationSettings,
    WebHookSettings,
)
from .typing import KeyOfSettings, KeySetOfSettings
from .helpers import update_settings, shadow_settings
from .setting_manager import SettingsManager


__all__ = (
    'DEFAULT_SETTINGS_PATH',

    'EnvSettings',
    'Settings',
    'SettingsIn',
    'SettingsOut',

    'KeyOfSettings',
    'KeySetOfSettings',

    'HeaderOptions',
    'HeaderSettings',
    'DanmakuOptions',
    'DanmakuSettings',
    'RecorderOptions',
    'RecorderSettings',
    'PostprocessingSettings',
    'PostprocessingOptions',

    'TaskOptions',
    'TaskSettings',
    'OutputSettings',
    'LoggingSettings',
    'SpaceSettings',
    'EmailSettings',
    'ServerchanSettings',
    'PushplusSettings',
    'NotifierSettings',
    'NotificationSettings',
    'EmailNotificationSettings',
    'ServerchanNotificationSettings',
    'PushplusNotificationSettings',
    'WebHookSettings',

    'update_settings',
    'shadow_settings',
    'SettingsManager',
)



from .notifiers import (
    Notifier,
    MessageNotifier,
    EmailNotifier,
    ServerchanNotifier,
    PushplusNotifier,
)
from .providers import (
    MessagingProvider,
    EmailService,
    Serverchan,
    Pushplus,
)


__all__ = (
    'MessagingProvider',
    'EmailService',
    'Serverchan',
    'Pushplus',

    'Notifier',
    'MessageNotifier',
    'EmailNotifier',
    'ServerchanNotifier',
    'PushplusNotifier',
)




from rx.subject import Subject
from rx.core import Observable

from .typing import Event
from ..utils.patterns import Singleton


__all__ = 'EventCenter',


class EventCenter(Singleton):
    def __init__(self) -> None:
        super().__init__()
        self._source = Subject()

    @property
    def events(self) -> Observable:
        return self._source

    def submit(self, event: Event) -> None:
        self._source.on_next(event)



from __future__ import annotations
import re
import html
import logging
from datetime import datetime
from typing import Iterator, Optional

import humanize

from .danmaku_receiver import DanmakuReceiver
from .danmaku_dumper import DanmakuDumper
from .stream_recorder import StreamRecorder
from ..event.event_emitter import EventListener, EventEmitter
from ..bili.live import Live
from ..bili.models import RoomInfo
from ..bili.danmaku_client import DanmakuClient
from ..bili.live_monitor import LiveMonitor, LiveEventListener
from ..bili.typing import QualityNumber
from ..utils.mixins import AsyncStoppableMixin


__all__ = 'RecorderEventListener', 'Recorder'


logger = logging.getLogger(__name__)


class RecorderEventListener(EventListener):
    async def on_recording_started(self, recorder: Recorder) -> None:
        ...

    async def on_recording_finished(self, recorder: Recorder) -> None:
        ...

    async def on_recording_cancelled(self, recorder: Recorder) -> None:
        ...


class Recorder(
    EventEmitter[RecorderEventListener], LiveEventListener, AsyncStoppableMixin
):
    def __init__(
        self,
        live: Live,
        danmaku_client: DanmakuClient,
        live_monitor: LiveMonitor,
        out_dir: str,
        path_template: str,
        *,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        danmu_uname: bool = False,
        filesize_limit: int = 0,
        duration_limit: int = 0,
    ) -> None:
        super().__init__()

        self._live = live
        self._danmaku_client = danmaku_client
        self._live_monitor = live_monitor

        self._recording: bool = False

        self._stream_recorder = StreamRecorder(
            self._live,
            out_dir=out_dir,
            path_template=path_template,
            buffer_size=buffer_size,
            read_timeout=read_timeout,
            filesize_limit=filesize_limit,
            duration_limit=duration_limit,
        )
        self._danmaku_receiver = DanmakuReceiver(danmaku_client)
        self._danmaku_dumper = DanmakuDumper(
            self._live,
            self._stream_recorder,
            self._danmaku_receiver,
            danmu_uname=danmu_uname,
        )

    @property
    def recording(self) -> bool:
        return self._recording

    @property
    def quality_number(self) -> QualityNumber:
        return self._stream_recorder.quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._stream_recorder.quality_number = value

    @property
    def real_quality_number(self) -> QualityNumber:
        return self._stream_recorder.real_quality_number

    @property
    def buffer_size(self) -> int:
        return self._stream_recorder.buffer_size

    @buffer_size.setter
    def buffer_size(self, value: int) -> None:
        self._stream_recorder.buffer_size = value

    @property
    def read_timeout(self) -> int:
        return self._stream_recorder.read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int) -> None:
        self._stream_recorder.read_timeout = value

    @property
    def danmu_uname(self) -> bool:
        return self._danmaku_dumper.danmu_uname

    @danmu_uname.setter
    def danmu_uname(self, value: bool) -> None:
        self._danmaku_dumper.danmu_uname = value

    @property
    def elapsed(self) -> float:
        return self._stream_recorder.elapsed

    @property
    def data_count(self) -> int:
        return self._stream_recorder.data_count

    @property
    def data_rate(self) -> float:
        return self._stream_recorder.data_rate

    @property
    def danmu_count(self) -> int:
        return self._danmaku_dumper.danmu_count

    @property
    def danmu_rate(self) -> float:
        return self._danmaku_dumper.danmu_rate

    @property
    def out_dir(self) -> str:
        return self._stream_recorder.out_dir

    @out_dir.setter
    def out_dir(self, value: str) -> None:
        self._stream_recorder.out_dir = value

    @property
    def path_template(self) -> str:
        return self._stream_recorder.path_template

    @path_template.setter
    def path_template(self, value: str) -> None:
        self._stream_recorder.path_template = value

    @property
    def filesize_limit(self) -> int:
        return self._stream_recorder.filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._stream_recorder.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        return self._stream_recorder.duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._stream_recorder.duration_limit = value

    async def _do_start(self) -> None:
        self._live_monitor.add_listener(self)
        logger.debug('Started recorder')

        self._print_live_info()
        if self._live.is_living():
            await self._start_recording(stream_available=True)
        else:
            self._print_waiting_message()

    async def _do_stop(self) -> None:
        await self._stop_recording()
        self._live_monitor.remove_listener(self)
        logger.debug('Stopped recorder')

    def get_video_files(self) -> Iterator[str]:
        yield from self._stream_recorder.get_files()

    def get_danmaku_files(self) -> Iterator[str]:
        yield from self._danmaku_dumper.get_files()

    async def on_live_began(self, live: Live) -> None:
        logger.info('The live has began')
        self._print_live_info()
        await self._start_recording()

    async def on_live_ended(self, live: Live) -> None:
        logger.info('The live has ended')
        await self._stop_recording()
        self._print_waiting_message()

    async def on_live_stream_available(self, live: Live) -> None:
        logger.debug('The live stream becomes available')
        await self._stream_recorder.start()

    async def on_live_stream_reset(self, live: Live) -> None:
        logger.warning('The live stream has been reset')

    async def on_room_changed(self, room_info: RoomInfo) -> None:
        self._print_changed_room_info(room_info)
        self._stream_recorder.update_progress_bar_info()

    async def _start_recording(self, stream_available: bool = False) -> None:
        if self._recording:
            return
        self._recording = True

        self._danmaku_dumper.enable()
        self._danmaku_receiver.start()
        await self._prepare()
        if stream_available:
            await self._stream_recorder.start()

        logger.info('Started recording')
        await self._emit('recording_started', self)

    async def _stop_recording(self) -> None:
        if not self._recording:
            return
        self._recording = False

        await self._stream_recorder.stop()
        self._danmaku_dumper.disable()
        self._danmaku_receiver.stop()

        if self._stopped:
            logger.info('Recording Cancelled')
            await self._emit('recording_cancelled', self)
        else:
            logger.info('Recording Finished')
            await self._emit('recording_finished', self)

    async def _prepare(self) -> None:
        live_start_time = self._live.room_info.live_start_time
        self._danmaku_dumper.set_live_start_time(live_start_time)
        self._danmaku_dumper.clear_files()
        self._stream_recorder.clear_files()

    def _print_waiting_message(self) -> None:
        logger.info('Waiting... until the live starts')

    def _print_live_info(self) -> None:
        room_info = self._live.room_info
        user_info = self._live.user_info

        if room_info.live_start_time > 0:
            live_start_time = str(
                datetime.fromtimestamp(room_info.live_start_time)
            )
        else:
            live_start_time = 'NULL'

        desc = re.sub(
            r'</?[a-zA-Z][a-zA-Z\d]*[^<>]*>',
            '',
            re.sub(r'<br\s*/?>', '\n', html.unescape(room_info.description))
        ).strip()

        msg = f"""
================================== User Info ==================================
user name        : {user_info.name}
gender           : {user_info.gender}
sign             : {user_info.sign}
uid              : {user_info.uid}
level            : {user_info.level}
---------------------------------- Room Info ----------------------------------
title            : {room_info.title}
cover            : {room_info.cover}
online           : {humanize.intcomma(room_info.online)}
live status      : {room_info.live_status.name}
live start time  : {live_start_time}
room id          : {room_info.room_id}
short room id    : {room_info.short_room_id or 'NULL'}
area id          : {room_info.area_id}
area name        : {room_info.area_name}
parent area id   : {room_info.parent_area_id}
parent area name : {room_info.parent_area_name}
tags             : {room_info.tags}
description      :
{desc}
===============================================================================
"""
        logger.info(msg)

    def _print_changed_room_info(self, room_info: RoomInfo) -> None:
        msg = f"""
================================= Room Change =================================
title            : {room_info.title}
area id          ：{room_info.area_id}
area name        : {room_info.area_name}
parent area id   : {room_info.parent_area_id}
parent area name : {room_info.parent_area_name}
===============================================================================
"""
        logger.info(msg)





import attr


@attr.s(auto_attribs=True, slots=True, frozen=True)
class Metadata:
    user_name: str
    room_id: int
    room_title: str
    area: str
    parent_area: str
    live_start_time: int  # seconds
    record_start_time: int  # seconds
    recorder: str


@attr.s(auto_attribs=True, slots=True, frozen=True)
class Danmu:
    stime: float
    mode: int
    size: int
    color: int
    date: int  # milliseconds
    pool: int
    uid: str  # uid hash
    dmid: int
    text: str



import logging
from asyncio import Queue, QueueFull
from typing import Final


from .models import DanmuMsg
from ..bili.danmaku_client import (
    DanmakuClient, DanmakuListener, DanmakuCommand
)
from ..bili.typing import Danmaku
from ..utils.mixins import StoppableMixin


__all__ = 'DanmakuReceiver',


logger = logging.getLogger(__name__)


class DanmakuReceiver(DanmakuListener, StoppableMixin):
    _MAX_QUEUE_SIZE: Final[int] = 2000

    def __init__(self, danmaku_client: DanmakuClient) -> None:
        super().__init__()
        self._danmaku_client = danmaku_client
        self._queue: Queue[DanmuMsg] = Queue(maxsize=self._MAX_QUEUE_SIZE)

    def _do_start(self) -> None:
        self._danmaku_client.add_listener(self)
        logger.debug('Started danmaku receiver')

    def _do_stop(self) -> None:
        self._danmaku_client.remove_listener(self)
        self._clear_queue()
        logger.debug('Stopped danmaku receiver')

    async def get_message(self) -> DanmuMsg:
        return await self._queue.get()

    async def on_danmaku_received(self, danmu: Danmaku) -> None:
        if danmu['cmd'] != DanmakuCommand.DANMU_MSG.value:
            return

        danmu_msg = DanmuMsg.from_cmd(danmu)
        try:
            self._queue.put_nowait(danmu_msg)
        except QueueFull:
            self._queue.get_nowait()  # discard the first item
            self._queue.put_nowait(danmu_msg)

    def _clear_queue(self) -> None:
        self._queue = Queue(maxsize=self._MAX_QUEUE_SIZE)



import traceback


def format_exception(exc: BaseException) -> str:
    exc_info = (type(exc), exc, exc.__traceback__)
    errmsg = 'Unhandled exception in event loop\n'
    errmsg += ''.join(traceback.format_exception(*exc_info))
    return errmsg




import attr


@attr.s(auto_attribs=True, frozen=True, slots=True)
class ApiRequestError(Exception):
    code: int
    message: str


class LiveRoomHidden(Exception):
    pass


class LiveRoomLocked(Exception):
    pass


class LiveRoomEncrypted(Exception):
    pass


class NoStreamUrlAvailable(Exception):
    pass



from .task_manager import RecordTaskManager
from .models import TaskStatus, TaskParam, FileDetail, RunningStatus, TaskData


__all__ = (
    'RecordTaskManager',

    'TaskStatus',
    'TaskParam',
    'TaskData',
    'FileDetail',
    'RunningStatus',
)



import logging
import asyncio
from abc import ABC, abstractmethod

from tenacity import (
    AsyncRetrying,
    wait_exponential,
    stop_after_delay,
    retry_if_exception,
)

from .providers import (
    EmailService,
    MessagingProvider,
    Serverchan,
    Pushplus,
)
from .message import (
    make_live_info_content,
    make_disk_usage_content,
    make_exception_content,
)
from ..utils.mixins import SwitchableMixin
from ..exception import ExceptionCenter
from ..event import (
    EventCenter,
    LiveBeganEvent,
    LiveEndedEvent,
    SpaceNoEnoughEvent,
)
from ..event.typing import Event


__all__ = (
    'Notifier',
    'MessageNotifier',
    'EmailNotifier',
    'ServerchanNotifier',
    'PushplusNotifier'
)


logger = logging.getLogger(__name__)


class Notifier(SwitchableMixin, ABC):
    def __init__(
        self,
        *,
        notify_began: bool = False,
        notify_ended: bool = False,
        notify_space: bool = False,
        notify_error: bool = False,
    ) -> None:
        super().__init__()
        self.notify_began = notify_began
        self.notify_ended = notify_ended
        self.notify_space = notify_space
        self.notify_error = notify_error

    def _do_enable(self) -> None:
        events = EventCenter.get_instance().events
        self._event_subscription = events.subscribe(self._on_event)
        exceptions = ExceptionCenter.get_instance().exceptions
        self._exc_subscription = exceptions.subscribe(self._on_exception)

    def _do_disable(self) -> None:
        self._event_subscription.dispose()
        self._exc_subscription.dispose()

    def _on_event(self, event: Event) -> None:
        if isinstance(event, LiveBeganEvent):
            if self._should_notify_live_began(event):
                self._notify_live_began(event)
        elif isinstance(event, LiveEndedEvent):
            if self._should_notify_live_ended(event):
                self._notify_live_ended(event)
        elif isinstance(event, SpaceNoEnoughEvent):
            if self._should_notify_space_no_enough(event):
                self._notify_space_no_enough(event)
        else:
            pass

    def _on_exception(self, exc: BaseException) -> None:
        if self._should_notify_exception(exc):
            self._notify_exception(exc)

    def _should_notify_live_began(self, event: LiveBeganEvent) -> bool:
        return self.notify_began

    def _should_notify_live_ended(self, event: LiveEndedEvent) -> bool:
        return self.notify_ended

    def _should_notify_space_no_enough(
        self, event: SpaceNoEnoughEvent
    ) -> bool:
        return self.notify_space

    def _should_notify_exception(self, exc: BaseException) -> bool:
        return self.notify_error

    @abstractmethod
    def _notify_live_began(self, event: LiveBeganEvent) -> None:
        raise NotImplementedError

    @abstractmethod
    def _notify_live_ended(self, event: LiveEndedEvent) -> None:
        raise NotImplementedError

    @abstractmethod
    def _notify_space_no_enough(self, event: SpaceNoEnoughEvent) -> None:
        raise NotImplementedError

    @abstractmethod
    def _notify_exception(self, exc: BaseException) -> None:
        raise NotImplementedError


class MessageNotifier(Notifier, ABC):
    provider: MessagingProvider

    def _notify_live_began(self, event: LiveBeganEvent) -> None:
        title = f'{event.data.user_info.name} 开播啦'
        content = make_live_info_content(
            event.data.user_info, event.data.room_info
        )
        self._send_message(title, content)

    def _notify_live_ended(self, event: LiveEndedEvent) -> None:
        title = f'{event.data.user_info.name} 下播了'
        content = make_live_info_content(
            event.data.user_info, event.data.room_info
        )
        self._send_message(title, content)

    def _notify_space_no_enough(self, event: SpaceNoEnoughEvent) -> None:
        title = '空间不足！'
        content = make_disk_usage_content(event.data)
        self._send_message(title, content)

    def _notify_exception(self, exc: BaseException) -> None:
        title = '出错了~'
        content = make_exception_content(exc)
        self._send_message(title, content)

    def _send_message(self, title: str, content: str) -> None:
        asyncio.create_task(self._send_message_async(title, content))

    async def _send_message_async(self, title: str, content: str) -> None:
        try:
            async for attempt in AsyncRetrying(
                reraise=True,
                stop=stop_after_delay(300),
                wait=wait_exponential(multiplier=0.1, max=10),
                retry=retry_if_exception(
                    lambda e: not isinstance(e, ValueError)
                ),
            ):
                with attempt:
                    await self.provider.send_message(title, content)
        except Exception as e:
            logger.warning('Failed to send a message via {}: {}'.format(
                self.provider.__class__.__name__, repr(e)
            ))


class EmailNotifier(MessageNotifier):
    provider = EmailService.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Email notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Email notifier')


class ServerchanNotifier(MessageNotifier):
    provider = Serverchan.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Serverchan notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Serverchan notifier')


class PushplusNotifier(MessageNotifier):
    provider = Pushplus.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Pushplus notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Pushplus notifier')



from __future__ import annotations
from typing import TYPE_CHECKING

from .event_center import EventCenter
from .models import (
    LiveBeganEvent,
    LiveBeganEventData,
    LiveEndedEvent,
    LiveEndedEventData,
    RoomChangeEvent,
    RoomChangeEventData,
    FileCompletedEvent,
    FileCompletedEventData,
    SpaceNoEnoughEvent,
    SpaceNoEnoughEventData,
)
from ..bili.live import Live
from ..bili.models import RoomInfo
from ..bili.live_monitor import LiveEventListener
from ..disk_space import SpaceEventListener
from ..postprocess import PostprocessorEventListener
if TYPE_CHECKING:
    from ..bili.live_monitor import LiveMonitor
    from ..disk_space import SpaceMonitor, DiskUsage
    from ..postprocess import Postprocessor


__all__ = (
    'LiveEventSubmitter',
    'SpaceEventSubmitter',
    'PostprocessorEventSubmitter',
)

event_center = EventCenter.get_instance()


class LiveEventSubmitter(LiveEventListener):
    def __init__(self, live_monitor: LiveMonitor) -> None:
        super().__init__()
        live_monitor.add_listener(self)

    async def on_live_began(self, live: Live) -> None:
        data = LiveBeganEventData(live.user_info, live.room_info)
        event_center.submit(LiveBeganEvent.from_data(data))

    async def on_live_ended(self, live: Live) -> None:
        data = LiveEndedEventData(live.user_info, live.room_info)
        event_center.submit(LiveEndedEvent.from_data(data))

    async def on_room_changed(self, room_info: RoomInfo) -> None:
        data = RoomChangeEventData(room_info)
        event_center.submit(RoomChangeEvent.from_data(data))


class SpaceEventSubmitter(SpaceEventListener):
    def __init__(self, space_monitor: SpaceMonitor) -> None:
        super().__init__()
        space_monitor.add_listener(self)

    async def on_space_no_enough(
        self, path: str, threshold: int, disk_usage: DiskUsage
    ) -> None:
        data = SpaceNoEnoughEventData(path, threshold, disk_usage)
        event_center.submit(SpaceNoEnoughEvent.from_data(data))


class PostprocessorEventSubmitter(PostprocessorEventListener):
    def __init__(self, postprocessor: Postprocessor) -> None:
        super().__init__()
        postprocessor.add_listener(self)

    async def on_file_completed(self, room_id: int, path: str) -> None:
        data = FileCompletedEventData(room_id, path)
        event_center.submit(FileCompletedEvent.from_data(data))



import logging
import asyncio
import smtplib
from abc import ABC, abstractmethod
from typing import TypedDict, cast, Literal
from email.message import EmailMessage
from http.client import HTTPException

import aiohttp

from ..utils.patterns import Singleton


__all__ = 'MessagingProvider', 'EmailService', 'Serverchan', 'Pushplus'


logger = logging.getLogger(__name__)


class MessagingProvider(Singleton, ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    async def send_message(self, title: str, content: str) -> None:
        ...


MSG_TYPE = Literal['plain', 'html']


class EmailService(MessagingProvider):
    def __init__(
        self,
        src_addr: str = '',
        dst_addr: str = '',
        auth_code: str = '',
        smtp_host: str = 'smtp.163.com',
        smtp_port: int = 465,
    ) -> None:
        super().__init__()
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.auth_code = auth_code
        self.smtp_host = smtp_host
        self.smtp_port = smtp_port

    async def send_message(
        self, subject: str, content: str, msg_type: MSG_TYPE = 'plain'
    ) -> None:
        self._check_parameters()
        await asyncio.get_running_loop().run_in_executor(
            None, self._send_email, subject, content, msg_type
        )

    def _send_email(
        self, subject: str, content: str, msg_type: MSG_TYPE = 'plain'
    ) -> None:
        msg = EmailMessage()
        msg['Subject'] = subject
        msg['From'] = self.src_addr
        msg['To'] = self.dst_addr
        msg.set_content(content, subtype=msg_type, charset='utf-8')

        with smtplib.SMTP_SSL(self.smtp_host, self.smtp_port) as smtp:
            # smtp.set_debuglevel(1)
            smtp.login(self.src_addr, self.auth_code)
            smtp.send_message(msg, self.src_addr, self.dst_addr)

    def _check_parameters(self) -> None:
        if not self.src_addr:
            raise ValueError('No source email address supplied')
        if not self.dst_addr:
            raise ValueError('No destination email address supplied')
        if not self.auth_code:
            raise ValueError('No auth code supplied')


class Serverchan(MessagingProvider):
    def __init__(self, sendkey: str = '') -> None:
        super().__init__()
        self.sendkey = sendkey

    async def send_message(self, title: str, content: str) -> None:
        self._check_parameters()
        await self._post_message(title, content)

    def _check_parameters(self) -> None:
        if not self.sendkey:
            raise ValueError('No sendkey supplied')

    async def _post_message(self, title: str, content: str) -> None:
        url = f'https://sctapi.ftqq.com/{self.sendkey}.send'
        payload = {'text': title, 'desp': content}

        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.post(url, data=payload):
                pass


class PushplusResponse(TypedDict):
    code: int
    msg: str
    data: str


class Pushplus(MessagingProvider):
    url = 'http://pushplus.hxtrip.com/send'

    def __init__(self, token: str = '', topic: str = '') -> None:
        super().__init__()
        self.token = token
        self.topic = topic

    async def send_message(self, title: str, content: str) -> None:
        self._check_parameters()
        await self._post_message(title, content)

    def _check_parameters(self) -> None:
        if not self.token:
            raise ValueError('No token supplied')

    async def _post_message(self, title: str, content: str) -> None:
        payload = {
            'title': title,
            'content': content,
            'token': self.token,
            'topic': self.topic,
            'template': 'html',
        }

        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.post(self.url, json=payload) as res:
                response = cast(PushplusResponse, await res.json())
                if response['code'] != 200:
                    raise HTTPException(response['code'], response['msg'])



import logging

import attr

from ..bili.typing import Danmaku


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class DanmuMsg:
    mode: int
    size: int  # font size
    color: int
    date: int  # a timestamp in miliseconds
    dmid: int
    pool: int
    uid: str  # midHash
    text: str
    uname: str  # sender name

    @staticmethod
    def from_cmd(danmu: Danmaku) -> 'DanmuMsg':
        info = danmu['info']
        return DanmuMsg(
            mode=int(info[0][1]),
            size=int(info[0][2]),
            color=int(info[0][3]),
            date=int(info[0][4]),
            dmid=int(info[0][5]),
            pool=int(info[0][6]),
            uid=info[0][7],
            text=info[1],
            uname=info[2][1],
        )



from __future__ import annotations
import io
import math
from abc import ABC, abstractmethod
import json
import logging
from typing import (
    Any, BinaryIO, Dict, List, Final, Iterable, Iterator, Optional, Tuple,
    Protocol, TypedDict, Union, cast
)

import attr
from rx.subject import Subject
from rx.core import Observable

from .models import FlvHeader, FlvTag, ScriptTag, VideoTag, AudioTag
from .data_analyser import DataAnalyser
from .limit_checker import LimitChecker
from .parameters_checker import ParametersChecker
from .io import FlvReader, FlvWriter
from .io_protocols import RandomIO
from .utils import format_offest, format_timestamp
from .exceptions import (
    FlvStreamCorruptedError,
    AudioParametersChanged,
    VideoParametersChanged,
    FileSizeOverLimit,
    DurationOverLimit,
)
from .common import (
    is_audio_tag, is_metadata_tag, is_video_tag, parse_metadata, rpeek_tags,
    enrich_metadata, update_metadata, is_data_tag, read_tags_in_duration,
    is_sequence_header
)
from ..path import extra_metadata_path


__all__ = 'StreamProcessor', 'BaseOutputFileManager', 'JoinPoint'


logger = logging.getLogger(__name__)


class StreamProcessor:
    # Number of tags for determining whether or not tags are duplicated
    _TAG_SEQUENCE_COUNT: Final[int] = 3
    # Max duration in milliseconds the duplicated tags might last
    _MAX_DURATION: Final[int] = 20000

    def __init__(
        self,
        file_manager: OutputFileManager,
        *,
        metadata: Optional[Dict[str, Any]] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
        disable_limit: bool = False,
        analyse_data: bool = False,
        dedup_join: bool = False,
        save_extra_metadata: bool = False,
    ) -> None:
        self._file_manager = file_manager
        self._parameters_checker = ParametersChecker()
        if not disable_limit:
            self._limit_checker = LimitChecker(filesize_limit, duration_limit)
        if analyse_data:
            self._data_analyser = DataAnalyser()

        self._metadata = metadata.copy() if metadata else {}

        self._disable_limit = disable_limit
        self._analyse_data = analyse_data
        self._dedup_join = dedup_join
        self._save_x_metadata = save_extra_metadata

        self._cancelled: bool = False
        self._finalized: bool = False
        self._stream_count: int = 0
        self._size_updates = Subject()
        self._time_updates = Subject()

        self._delta: int = 0
        self._has_audio: bool = False
        self._metadata_tag: ScriptTag
        self._last_ts: int = 0
        self._last_tags: List[FlvTag] = []
        self._join_points: List[JoinPoint] = []

    @property
    def filesize_limit(self) -> int:
        if self._disable_limit:
            return 0
        return self._limit_checker.filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        if not self._disable_limit:
            self._limit_checker.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        if self._disable_limit:
            return 0
        return self._limit_checker.duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        if not self._disable_limit:
            self._limit_checker.duration_limit = value

    @property
    def join_points(self) -> Iterator[JoinPoint]:
        for point in self._join_points:
            yield point

    @property
    def size_updates(self) -> Observable:
        return self._size_updates

    @property
    def time_updates(self) -> Observable:
        return self._time_updates

    @property
    def cancelled(self) -> bool:
        return self._cancelled

    @property
    def finalized(self) -> bool:
        return self._finalized

    def cancel(self) -> None:
        self._cancelled = True

    def process_stream(self, stream: RandomIO) -> None:
        assert not self._cancelled and not self._finalized, \
            'should not be called after the processing cancelled or finalized'
        self._stream_count += 1
        self._process_stream(stream)

    def finalize(self) -> None:
        assert not self._finalized, \
            'should not be called after the processing finalized'
        self._finalized = True

        if not self._need_to_finalize():
            logger.debug('No need to finalize stream processing')
            return

        self._complete_file()
        logger.debug('Finalized stream processing')

    def _need_to_finalize(self) -> bool:
        return self._stream_count > 0 and len(self._last_tags) > 0

    def _new_file(self) -> None:
        if not self._disable_limit:
            self._limit_checker.reset()

        self._out_file = self._file_manager.create_file()
        self._out_reader = FlvReader(self._out_file)
        self._out_writer = FlvWriter(self._out_file)

        logger.debug(f'new file: {self._file_manager.curr_path}')

    def _complete_file(self) -> None:
        curr_path = self._file_manager.curr_path

        if self._save_x_metadata:
            self._save_extra_metadata()

        self._update_metadata_tag()
        self._file_manager.close_file()

        if self._analyse_data:
            self._data_analyser.reset()

        logger.debug(f'complete file: {curr_path}')

    def _discard_file(self) -> None:
        curr_path = self._file_manager.curr_path

        self._file_manager.close_file()

        if self._analyse_data:
            self._data_analyser.reset()

        logger.debug(f'discard file: {curr_path}')

    def _reset(self) -> None:
        self._discard_file()
        self._stream_count = 0
        logger.debug('Reset stream processing')

    def _process_stream(self, stream: RandomIO) -> None:
        logger.debug(f'Processing the {self._stream_count}th stream...')

        self._in_reader = FlvReaderWithTimestampFix(stream)
        flv_header = self._read_header()
        self._has_audio = flv_header.has_audio()
        assert flv_header.has_video(), 'no video in the stream!'

        try:
            first_data_tag = self._read_first_data_tag()
            if self._stream_count == 1:
                self._process_initial_stream(flv_header, first_data_tag)
            else:
                self._process_subsequent_stream(first_data_tag)
        except (
            AudioParametersChanged, VideoParametersChanged,
            FileSizeOverLimit, DurationOverLimit,
        ):
            self._process_split_stream(flv_header)

        logger.debug(f'Completed processing the {self._stream_count}th stream')

    def _process_initial_stream(
        self, flv_header: FlvHeader, first_data_tag: FlvTag
    ) -> None:
        self._new_file()

        try:
            self._write_header(flv_header)
            self._transfer_meta_tags()
            self._transfer_first_data_tag(first_data_tag)
        except Exception:
            self._reset()
            raise
        del first_data_tag
        self._transfer_tags_until_complete()

    def _process_subsequent_stream(self, first_data_tag: FlvTag) -> None:
        tags: List[FlvTag] = []

        if self._dedup_join:
            tags, exc = self._read_tags_for_deduplication()

            if (index := self._find_last_duplicated_tag(tags)) >= 0:
                seamless = True
                self._delta = self._calc_delta_duplicated(tags[index])
                tags = tags[index + 1:]
                if not tags:
                    tags = [self._read_first_data_tag()]

        if not self._dedup_join or index == -1:
            seamless = False
            self._delta = self._calc_delta_no_duplicated(first_data_tag)
            tags.insert(0, first_data_tag)

        offset = self._out_file.tell()
        timestamp = tags[0].timestamp + self._delta
        self._add_join_point(offset, timestamp, seamless)

        self._transfer_tags(tags)
        del first_data_tag, tags

        if self._dedup_join and exc:
            raise exc

        self._transfer_tags_until_complete()

    def _process_split_stream(self, flv_header: FlvHeader) -> None:
        self._complete_file()

        first_data_tag: FlvTag
        if (
            not self._disable_limit and (
                self._limit_checker.is_filesize_over_limit() or
                self._limit_checker.is_duration_over_limit()
            )
        ):
            assert self._limit_checker.last_keyframe_tag is not None
            last_keyframe_tag = self._limit_checker.last_keyframe_tag
            original_ts = last_keyframe_tag.timestamp - self._delta
            first_data_tag = last_keyframe_tag.evolve(timestamp=original_ts)
        else:
            first_data_tag = self._read_first_data_tag()

        try:
            self._process_initial_stream(flv_header, first_data_tag)
        except (
            AudioParametersChanged, VideoParametersChanged,
            FileSizeOverLimit, DurationOverLimit,
        ):
            self._process_split_stream(flv_header)

    def _transfer_meta_tags(self) -> None:
        logger.debug('Transfering meta tags...')

        if self._parameters_checker.last_metadata_tag is None:
            raise FlvStreamCorruptedError('No metadata tag in the stream')
        if self._parameters_checker.last_video_header_tag is None:
            raise FlvStreamCorruptedError('No video header tag in the stream')
        if self._has_audio:
            if self._parameters_checker.last_audio_header_tag is None:
                raise FlvStreamCorruptedError(
                    'No audio header tag in the stream'
                )
        metadata_tag = self._parameters_checker.last_metadata_tag
        video_header_tag = self._parameters_checker.last_video_header_tag
        audio_header_tag = self._parameters_checker.last_audio_header_tag

        offset = self._out_file.tell()
        self._metadata_tag = self._enrich_metadata(metadata_tag, offset)

        offset_delta = self._metadata_tag.tag_size - metadata_tag.tag_size
        self._update_injected_metadata(offset_delta)

        self._write_tag(
            self._correct_ts(self._metadata_tag, -self._metadata_tag.timestamp)
        )
        self._write_tag(
            self._correct_ts(video_header_tag, -video_header_tag.timestamp)
        )
        if audio_header_tag is not None:
            self._write_tag(
                self._correct_ts(audio_header_tag, -audio_header_tag.timestamp)
            )

        logger.debug('Meta tags have been transfered')

    def _transfer_first_data_tag(self, tag: FlvTag) -> None:
        logger.debug(f'Transfer the first data tag: {tag}')
        self._delta = -tag.timestamp
        self._transfer_tags([tag])
        logger.debug('The first data tag has been transfered')

    def _read_tags_for_deduplication(
        self
    ) -> Tuple[List[FlvTag], Optional[Exception]]:
        logger.debug('Reading tags for tag deduplication...')
        tags: List[FlvTag] = []
        try:
            for tag in read_tags_in_duration(
                self._in_reader, self._MAX_DURATION
            ):
                tags.append(tag)
        except Exception as exc:
            logger.debug(f'Failed to read data, due to: {repr(exc)}')
            return tags, exc
        else:
            return tags, None
        finally:
            logger.debug('Read {} tags, total size: {}'.format(
                len(tags), sum(t.tag_size for t in tags)
            ))

    def _transfer_tags_until_complete(self) -> None:
        logger.debug('Transfering tags until complete...')
        self._transfer_tags(self._read_tags_from_in_stream())

    def _read_tags_from_in_stream(self) -> Iterator[FlvTag]:
        while not self._cancelled:
            try:
                tag = self._in_reader.read_tag()
                self._parameters_checker.check_tag(tag)
                yield tag
            except EOFError:
                logger.debug('The input stream exhausted')
                break
            except Exception as e:
                logger.debug(f'Failed to read data, due to: {repr(e)}')
                raise
        else:
            logger.debug('Cancelled reading tags from the input stream')

    def _transfer_tags(self, tags: Iterable[FlvTag]) -> None:
        logger.debug(f'Transfering tags... timestamp delta: {self._delta}')

        try:
            count: int = 0
            for tag in filter(lambda t: not is_sequence_header(t), tags):
                self._ensure_ts_correct(tag)
                self._write_tag(self._correct_ts(tag, self._delta))
                count += 1
        finally:
            logger.debug(f'{count} tags have been transfered')

            if count > 0:
                self._update_last_out_tags()

    def _add_join_point(
        self, offset: int, timestamp: int, seamless: bool
    ) -> None:
        join_point = JoinPoint(offset, timestamp, seamless)
        self._join_points.append(join_point)
        logger.debug(f'{repr(join_point)}; {join_point}')

    def _update_last_out_tags(self) -> None:
        self._last_tags = list(rpeek_tags(
            self._out_file, self._out_reader, self._TAG_SEQUENCE_COUNT
        ))

    def _find_last_duplicated_tag(self, tags: List[FlvTag]) -> int:
        logger.debug('Finding duplicated tags...')

        last_out_tag = self._last_tags[0]
        logger.debug(f'The last output tag is {last_out_tag}')

        for idx, tag in enumerate(tags):
            if not tag.is_the_same_as(last_out_tag):
                continue

            if not all(
                map(
                    lambda t: t[0].is_the_same_as(t[1]),
                    zip(reversed(tags[:idx]), self._last_tags[1:])
                )
            ):
                continue

            logger.debug(f'The last duplicated tag found at {idx} is {tag}')
            return idx

        logger.debug('No duplicated tags found')
        return -1

    def _read_header(self) -> FlvHeader:
        try:
            return self._in_reader.read_header()
        except Exception as exc:
            raise FlvStreamCorruptedError(repr(exc))

    def _read_first_data_tag(self) -> Union[AudioTag, VideoTag]:
        for tag in self._read_tags_from_in_stream():
            if is_data_tag(tag):
                return tag
        raise FlvStreamCorruptedError('No data tag found in the stream!')

    def _read_tags(self, count: int) -> Iterator[FlvTag]:
        assert count > 0, 'count must greater than 0'
        for c, tag in enumerate(self._read_tags_from_in_stream(), start=1):
            yield tag
            if c >= count:
                break

    def _write_header(self, header: FlvHeader) -> None:
        try:
            size = self._out_writer.write_header(header)
            logger.debug('The flv header has been copied')
        except Exception as exc:
            logger.debug(f'Failed to write data, due to: {repr(exc)}')
            raise

        if not self._disable_limit:
            self._limit_checker.check_header(header)
        if self._analyse_data:
            self._data_analyser.analyse_header(header)
        self._size_updates.on_next(size)

    def _write_tag(self, tag: FlvTag) -> None:
        if self._analyse_data:
            offset = self._out_file.tell()
            tag = tag.evolve(offset=offset)

        try:
            size = self._out_writer.write_tag(tag)
        except Exception as exc:
            logger.debug(f'Failed to write data, due to: {repr(exc)}')
            raise

        self._last_ts = tag.timestamp

        if not self._disable_limit:
            self._limit_checker.check_tag(tag)

        if self._analyse_data:
            self._data_analyser.analyse_tag(tag)

        self._size_updates.on_next(size)
        self._time_updates.on_next(tag.timestamp)

    def _ensure_ts_correct(self, tag: FlvTag) -> None:
        if tag.timestamp + self._delta < 0:
            self._delta = -tag.timestamp
            logger.warning('Incorrect timestamp: {}, new delta: {}'.format(
                tag, self._delta
            ))

    def _correct_ts(self, tag: FlvTag, delta: int) -> FlvTag:
        if delta == 0:
            return tag
        return tag.evolve(timestamp=tag.timestamp + delta)

    def _calc_delta_duplicated(self, last_duplicated_tag: FlvTag) -> int:
        return self._last_tags[0].timestamp - last_duplicated_tag.timestamp

    def _calc_delta_no_duplicated(self, first_data_tag: FlvTag) -> int:
        return (
            self._last_tags[0].timestamp - first_data_tag.timestamp +
            self._in_reader.calc_interval(first_data_tag)
        )

    def _enrich_metadata(
        self, old_metadata_tag: ScriptTag, offset: int
    ) -> ScriptTag:
        # ensure the duration and filesize property exists in the metadata and
        # init them.
        self._metadata.update({'duration': 0.0, 'filesize': 0.0})
        # merge the metadata into the metadata tag
        return enrich_metadata(old_metadata_tag, self._metadata, offset)

    def _update_injected_metadata(self, offset_delta: int) -> None:
        updated = False

        if (keyframes := self._metadata.get('keyframes')):
            keyframes['filepositions'] = list(
                map(lambda p: p + offset_delta, keyframes['filepositions'])
            )
            if 'lastkeyframelocation' in self._metadata:
                self._metadata['lastkeyframelocation'] = \
                    keyframes['filepositions'][-1]
            updated = True

        if (join_points := self._metadata.get('joinpoints')):
            join_points = cast(List[JoinPointData], join_points)
            self._metadata['joinpoints'] = list(
                {**p, 'offset': p['offset'] + offset_delta}
                for p in join_points
            )
            updated = True

        if updated:
            self._metadata_tag = \
                update_metadata(self._metadata_tag, self._metadata)

    def _update_metadata_tag(self) -> None:
        last_tag = self._last_tags[0]
        duration = last_tag.timestamp / 1000
        filesize = float(last_tag.next_tag_offset)
        updates = dict(duration=duration, filesize=filesize)
        self._metadata_tag = update_metadata(self._metadata_tag, updates)
        self._out_file.seek(self._metadata_tag.offset)
        self._out_writer.write_tag(self._metadata_tag)
        logger.debug('The metadata tag has been updated')

    def _save_extra_metadata(self) -> None:
        if self._analyse_data:
            metadata = attr.asdict(
                self._data_analyser.make_metadata(),
                filter=lambda a, v: v is not None,
            )
        else:
            metadata = {}

        metadata['joinpoints'] = list(
            map(lambda p: p.to_metadata_value(), self._join_points)
        )

        path = extra_metadata_path(self._file_manager.curr_path)
        with open(path, 'wt', encoding='utf8') as file:
            json.dump(metadata, file)

        logger.debug('The extra metadata has been saved')


@attr.s(auto_attribs=True, slots=True, frozen=True)
class JoinPoint:
    offset: int
    timestamp: int  # milliseconds
    seamless: bool

    @classmethod
    def from_metadata_value(cls, value: JoinPointData) -> JoinPoint:
        return cls(
            offset=int(value['offset']),
            timestamp=int(value['timestamp']),
            seamless=value['seamless'],
        )

    def to_metadata_value(self) -> JoinPointData:
        return dict(
            offset=float(self.offset),
            timestamp=float(self.timestamp),
            seamless=self.seamless,
        )

    def __str__(self) -> str:
        return 'offset: {}, timestamp: {}, seamless: {}'.format(
            format_offest(self.offset),
            format_timestamp(self.timestamp),
            'yes' if self.seamless else 'no',
        )


class JoinPointData(TypedDict):
    offset: float
    timestamp: float
    seamless: bool


class OutputFileManager(Protocol):
    @property
    def curr_path(self) -> str:
        ...

    @property
    def curr_file(self) -> Optional[BinaryIO]:
        ...

    def create_file(self) -> BinaryIO:
        ...

    def close_file(self) -> None:
        ...


class BaseOutputFileManager(ABC):
    def __init__(self, buffer_size: Optional[int] = None) -> None:
        super().__init__()
        self.buffer_size = buffer_size or io.DEFAULT_BUFFER_SIZE  # bytes
        self._paths: List[str] = []
        self._curr_path = ''
        self._curr_file: Optional[BinaryIO] = None

    @property
    def curr_path(self) -> str:
        return self._curr_path

    @property
    def curr_file(self) -> Optional[BinaryIO]:
        return self._curr_file

    def has_file(self) -> bool:
        return len(self._paths) > 1

    def get_files(self) -> Iterator[str]:
        for file in self._paths:
            yield file

    def clear_files(self) -> None:
        self._paths.clear()

    def create_file(self) -> BinaryIO:
        assert self._curr_file is None
        path = self._make_path()
        file = open(path, mode='w+b', buffering=self.buffer_size)
        self._paths.append(path)
        self._curr_path = path
        self._curr_file = file
        return file

    def close_file(self) -> None:
        assert self._curr_file is not None
        self._curr_file.close()
        self._curr_path = ''
        self._curr_file = None

    @abstractmethod
    def _make_path(self) -> str:
        ...


class FlvReaderWithTimestampFix(FlvReader):
    def __init__(self, stream: RandomIO) -> None:
        super().__init__(stream)
        self._last_tag: Optional[FlvTag] = None
        self._last_video_tag: Optional[VideoTag] = None
        self._last_audio_tag: Optional[AudioTag] = None
        self._delta = 0
        # 15 is probably the minimal frame rate
        self._frame_rate = 15.0
        self._video_frame_interval = math.ceil(1000 / self._frame_rate)
        # AAC SoundRate always is 44 KHz
        self._sound_sample_interval = math.ceil(1000 / 44)

    @property
    def frame_rate(self) -> float:
        return self._frame_rate

    @property
    def sound_sample_interval(self) -> int:
        return self._sound_sample_interval

    @property
    def video_frame_interval(self) -> int:
        return self._video_frame_interval

    def read_tag(self, *, no_body: bool = False) -> FlvTag:
        while True:
            tag = super().read_tag(no_body=no_body)

            if self._last_tag is None:
                if is_data_tag(tag):
                    self._update_last_tags(tag)
                elif is_metadata_tag(tag):
                    self._update_parameters(tag)
                return tag

            if not is_data_tag(tag):
                return tag

            if self._is_ts_rebounded(tag):
                self._update_delta(tag)
                logger.warning(
                    f'Timestamp rebounded, updated delta: {self._delta}\n'
                    f'last video tag: {self._last_video_tag}\n'
                    f'last audio tag: {self._last_audio_tag}\n'
                    f'current tag: {tag}'
                )

            if self._is_ts_jumped(tag):
                self._update_delta(tag)
                logger.warning(
                    f'Timestamp jumped, updated delta: {self._delta}\n'
                    f'last tag: {self._last_tag}\n'
                    f'current tag: {tag}'
                )

            self._update_last_tags(tag)

            return self._correct_ts(tag)

    def rread_tag(self, *, no_body: bool = False) -> FlvTag:
        raise NotImplementedError()

    def read_body(self, tag: FlvTag) -> bytes:
        raise NotImplementedError()

    def calc_interval(self, tag: FlvTag) -> int:
        if is_audio_tag(tag):
            return self._sound_sample_interval
        elif is_video_tag(tag):
            return self._video_frame_interval
        else:
            logger.warning(f'Unexpected tag type: {tag}')
            return min(self._sound_sample_interval, self._video_frame_interval)

    def _is_ts_rebounded(self, tag: FlvTag) -> bool:
        if is_video_tag(tag):
            if self._last_video_tag is None:
                return False
            return tag.timestamp < self._last_video_tag.timestamp
        elif is_audio_tag(tag):
            if self._last_audio_tag is None:
                return False
            return tag.timestamp < self._last_audio_tag.timestamp
        else:
            return False

    def _is_ts_jumped(self, tag: FlvTag) -> bool:
        assert self._last_tag is not None
        return (
            tag.timestamp - self._last_tag.timestamp >
            max(self._sound_sample_interval, self._video_frame_interval)
        )

    def _update_last_tags(self, tag: FlvTag) -> None:
        self._last_tag = tag
        if is_video_tag(tag):
            self._last_video_tag = tag
        elif is_audio_tag(tag):
            self._last_audio_tag = tag

    def _update_parameters(self, tag: ScriptTag) -> None:
        metadata = parse_metadata(tag)
        frame_rate = metadata.get('fps') or metadata.get('framerate')

        if not frame_rate:
            return

        self._frame_rate = frame_rate
        self._video_frame_interval = math.ceil(1000 / frame_rate)

        logger.debug('frame rate: {}, video frame interval: {}'.format(
            frame_rate, self._video_frame_interval
        ))

    def _update_delta(self, tag: FlvTag) -> None:
        assert self._last_tag is not None
        self._delta = (
            self._last_tag.timestamp + self._delta - tag.timestamp +
            self.calc_interval(tag)
        )

    def _correct_ts(self, tag: FlvTag) -> FlvTag:
        if self._delta == 0:
            return tag
        return tag.evolve(timestamp=tag.timestamp + self._delta)



import os
import re
import shlex
from subprocess import Popen, PIPE
from typing import Final, List, Match, Optional, Union

import attr
from rx import create, operators as op
from rx.subject import Subject
from rx.core import Observable
from rx.core.typing import Observer, Scheduler, Disposable
from rx.scheduler.currentthreadscheduler import CurrentThreadScheduler
from tqdm import tqdm


__all__ = 'VideoRemuxer', 'RemuxResult', 'remux_video'


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RemuxProgress:
    time: int
    duration: int


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RemuxResult:
    return_code: int
    output: str

    def is_done(self) -> bool:
        return self.return_code == 0

    def is_successful(self) -> bool:
        return self.is_done() and not self.may_timestamps_incorrect()

    def is_warned(self) -> bool:
        return self.is_done() and self.may_timestamps_incorrect()

    def is_failed(self) -> bool:
        return not self.is_done()

    def may_timestamps_incorrect(self) -> bool:
        return 'Non-monotonous DTS in output stream' in self.output


class VideoRemuxer:
    _TIME_PATTERN: Final = re.compile(
        r'(?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+).(?P<millisecond>\d+)'
    )

    def __init__(self) -> None:
        self._duration: int = 0
        self._progress_updates = Subject()

    @property
    def progress_updates(self) -> Observable:
        return self._progress_updates

    def remux(
        self,
        in_path: str,
        out_path: str,
        metadata_path: Optional[str] = None,
    ) -> RemuxResult:
        if metadata_path is not None:
            cmd = f'ffmpeg -i "{in_path}" -i "{metadata_path}" ' \
                f'-map_metadata 1 -codec copy "{out_path}" -y'
        else:
            cmd = f'ffmpeg -i "{in_path}" -codec copy "{out_path}" -y'

        args = shlex.split(cmd)
        out_lines: List[str] = []

        with Popen(
            args, stderr=PIPE, encoding='utf8', errors='backslashreplace'
        ) as process:
            assert process.stderr is not None
            while True:
                line = process.stderr.readline()
                if not line:
                    if process.poll() is not None:
                        break
                    else:
                        continue
                self._parse_line(line)
                if self._should_output_line(line):
                    out_lines.append(line)

        if process.returncode == 0:
            self._complete_progress_updates()

        return RemuxResult(process.returncode, ''.join(out_lines))

    def _parse_line(self, line: str) -> None:
        line = line.strip()
        if line.startswith('frame='):
            self._parse_time(line)
        elif line.startswith('Duration:'):
            self._parse_duration(line)
        else:
            pass

    def _parse_duration(self, line: str) -> None:
        match = self._TIME_PATTERN.search(line)
        if match is not None:
            self._duration = self._calc_time(match)
            self._progress_updates.on_next(RemuxProgress(0, self._duration))

    def _parse_time(self, line: str) -> None:
        match = self._TIME_PATTERN.search(line)
        if match is not None:
            progress = RemuxProgress(self._calc_time(match), self._duration)
            self._progress_updates.on_next(progress)

    def _complete_progress_updates(self) -> None:
        progress = RemuxProgress(self._duration, self._duration)
        self._progress_updates.on_next(progress)
        self._progress_updates.on_completed()

    def _calc_time(self, match: Match[str]) -> int:
        result = match.groupdict()
        return (
            int(result['hour']) * 60 * 60 * 1000 +
            int(result['minute']) * 60 * 1000 +
            int(result['second']) * 1000 +
            int(result['millisecond'])
        )

    def _should_output_line(self, line: str) -> bool:
        line = line.strip()
        return not (
            line.startswith('frame=') or
            line.startswith('Press [q]')
        )


def remux_video(
    in_path: str,
    out_path: str,
    metadata_path: Optional[str] = None,
    *,
    report_progress: bool = False,
) -> Observable:
    def subscribe(
        observer: Observer[Union[RemuxProgress, RemuxResult]],
        scheduler: Optional[Scheduler] = None,
    ) -> Disposable:
        _scheduler = scheduler or CurrentThreadScheduler.singleton()

        def action(scheduler, state):  # type: ignore
            remuxer = VideoRemuxer()
            file_name = os.path.basename(in_path)

            with tqdm(desc='Remuxing', unit='ms', postfix=file_name) as pbar:
                def reset(progress: RemuxProgress) -> None:
                    pbar.reset(progress.duration)

                def update(progress: RemuxProgress) -> None:
                    pbar.update(progress.time - pbar.n)

                remuxer.progress_updates.pipe(op.first()).subscribe(reset)
                remuxer.progress_updates.pipe(op.skip(1)).subscribe(update)

                if report_progress:
                    remuxer.progress_updates.subscribe(
                        lambda p: observer.on_next(p)
                    )

                try:
                    result = remuxer.remux(in_path, out_path, metadata_path)
                except Exception as e:
                    observer.on_error(e)
                else:
                    observer.on_next(result)
                    observer.on_completed()

        return _scheduler.schedule(action)

    return create(subscribe)



from .postprocessor import Postprocessor, PostprocessorEventListener
from .models import ProcessStatus, DeleteStrategy


__all__ = (
    'Postprocessor',
    'PostprocessorEventListener',
    'ProcessStatus',
    'DeleteStrategy',
)




import humanize

from ..event import SpaceNoEnoughEventData

from ..bili.models import UserInfo, RoomInfo
from ..exception import format_exception


live_info_template = """
主播: {user_name}

标题: {room_title}

分区: {area_name}

房间: {room_id}
"""

disk_usage_template = """
路径: {path}

阈值: {threshold}

硬盘容量: {usage_total}

已用空间: {usage_used}

可用空间: {usage_free}
"""

exception_template = """
异常信息：

{exception_message}
"""


def make_live_info_content(user_info: UserInfo, room_info: RoomInfo) -> str:
    return live_info_template.format(
        user_name=user_info.name,
        room_title=room_info.title,
        area_name=room_info.area_name,
        room_id=room_info.room_id,
    )


def make_disk_usage_content(data: SpaceNoEnoughEventData) -> str:
    return disk_usage_template.format(
        path=data.path,
        threshold=humanize.naturalsize(data.threshold),
        usage_total=humanize.naturalsize(data.usage.total),
        usage_used=humanize.naturalsize(data.usage.used),
        usage_free=humanize.naturalsize(data.usage.free),
    )


def make_exception_content(exc: BaseException) -> str:
    return exception_template.format(
        exception_message=format_exception(exc),
    )



#!/usr/bin/env python3
import json
import struct
import zlib
import asyncio
import logging
from enum import IntEnum, Enum
from contextlib import suppress
from typing import Any, Dict, Final, Tuple, List, Union, cast, Optional

import aiohttp
from aiohttp import ClientSession
from tenacity import (
    retry,
    wait_exponential,
    retry_if_exception_type,
)

from .typing import Danmaku
from ..event.event_emitter import EventListener, EventEmitter
from ..exception import exception_callback
from ..utils.mixins import AsyncStoppableMixin
from ..logging.room_id import aio_task_with_room_id


__all__ = 'DanmakuClient', 'DanmakuListener', 'Danmaku', 'DanmakuCommand'


logger = logging.getLogger(__name__)


class DanmakuListener(EventListener):
    async def on_client_connected(self) -> None:
        ...

    async def on_client_disconnected(self) -> None:
        ...

    async def on_client_reconnected(self) -> None:
        ...

    async def on_danmaku_received(self, danmu: Danmaku) -> None:
        ...

    async def on_error_occurred(self, error: Exception) -> None:
        ...


class DanmakuClient(EventEmitter[DanmakuListener], AsyncStoppableMixin):
    _URL: Final[str] = 'wss://broadcastlv.chat.bilibili.com:443/sub'
    _HEARTBEAT_INTERVAL: Final[int] = 30

    def __init__(
        self,
        session: ClientSession,
        room_id: int,
        *,
        max_retries: int = 10,
    ) -> None:
        super().__init__()
        self.session = session
        self._room_id = room_id

        self._retry_delay: int = 0
        self._MAX_RETRIES: Final[int] = max_retries

    async def _do_start(self) -> None:
        await self._connect()
        await self._create_message_loop()
        logger.debug('Started danmaku client')

    async def _do_stop(self) -> None:
        await self._terminate_message_loop()
        await self._disconnect()
        logger.debug('Stopped danmaku client')

    async def reconnect(self) -> None:
        if self.stopped:
            return

        logger.debug('Reconnecting...')
        await self._disconnect()
        await self._connect()
        await self._emit('client_reconnected')

    @retry(
        wait=wait_exponential(multiplier=0.1, max=60),
        retry=retry_if_exception_type((
            asyncio.TimeoutError, aiohttp.ClientError,
        )),
    )
    async def _connect(self) -> None:
        logger.debug('Connecting to server...')
        await self._connect_websocket()
        await self._send_auth()
        reply = await self._recieve_auth_reply()
        self._handle_auth_reply(reply)
        logger.debug('Connected to server')
        await self._emit('client_connected')

    async def _connect_websocket(self) -> None:
        self._ws = await self.session.ws_connect(self._URL, timeout=5)
        logger.debug('Established WebSocket connection')

    async def _send_auth(self) -> None:
        auth_msg = json.dumps({
            'uid': 0,
            'roomid': self._room_id,  # must not be the short id!
            # 'protover': WS.BODY_PROTOCOL_VERSION_NORMAL,
            'protover': WS.BODY_PROTOCOL_VERSION_DEFLATE,
            'platform': 'web',
            'clientver': '1.1.031',
            'type': 2
        })
        data = Frame.encode(WS.OP_USER_AUTHENTICATION, auth_msg)
        await self._ws.send_bytes(data)
        logger.debug('Sent user authentication')

    async def _recieve_auth_reply(self) -> aiohttp.WSMessage:
        msg = await self._ws.receive(timeout=5)
        if msg.type != aiohttp.WSMsgType.BINARY:
            raise aiohttp.ClientError(msg)
        logger.debug('Recieved reply')
        return msg

    def _handle_auth_reply(self, reply: aiohttp.WSMessage) -> None:
        op, msg = Frame.decode(reply.data)
        assert op == WS.OP_CONNECT_SUCCESS
        msg = cast(str, msg)
        code = cast(int, json.loads(msg)['code'])

        if code == WS.AUTH_OK:
            logger.debug('Auth OK')
            self._create_heartbeat_task()
        elif code == WS.AUTH_TOKEN_ERROR:
            logger.debug('Auth Token Error')
            raise ValueError(f'Auth Token Error: {code}')
        else:
            raise ValueError(f'Unexpected code: {code}')

    async def _disconnect(self) -> None:
        await self._cancel_heartbeat_task()
        await self._close_websocket()
        logger.debug('Disconnected from server')
        await self._emit('client_disconnected')

    async def _close_websocket(self) -> None:
        with suppress(BaseException):
            await self._ws.close()

    def _create_heartbeat_task(self) -> None:
        self._heartbeat_task = asyncio.create_task(self._send_heartbeat())
        self._heartbeat_task.add_done_callback(exception_callback)

    async def _cancel_heartbeat_task(self) -> None:
        self._heartbeat_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._heartbeat_task

    @aio_task_with_room_id
    async def _send_heartbeat(self) -> None:
        data = Frame.encode(WS.OP_HEARTBEAT, '')
        while True:
            await self._ws.send_bytes(data)
            await asyncio.sleep(self._HEARTBEAT_INTERVAL)

    async def _create_message_loop(self) -> None:
        self._message_loop_task = asyncio.create_task(self._message_loop())
        self._message_loop_task.add_done_callback(exception_callback)
        logger.debug('Created message loop')

    async def _terminate_message_loop(self) -> None:
        self._message_loop_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._message_loop_task
        logger.debug('Terminated message loop')

    @aio_task_with_room_id
    async def _message_loop(self) -> None:
        while True:
            for msg in await self._receive():
                await self._dispatch_message(msg)

    async def _dispatch_message(self, msg: Dict[str, Any]) -> None:
        await self._emit('danmaku_received', msg)

    @retry(retry=retry_if_exception_type((asyncio.TimeoutError,)))
    async def _receive(self) -> List[Dict[str, Any]]:
        self._retry_count = 0
        self._retry_delay = 0

        while True:
            wsmsg = await self._ws.receive(timeout=self._HEARTBEAT_INTERVAL)

            if wsmsg.type == aiohttp.WSMsgType.BINARY:
                if (result := await self._handle_data(wsmsg.data)):
                    return result
            elif wsmsg.type == aiohttp.WSMsgType.ERROR:
                await self._handle_error(cast(Exception, wsmsg.data))
            elif wsmsg.type == aiohttp.WSMsgType.CLOSED:
                msg = 'WebSocket Closed'
                exc = aiohttp.WebSocketError(self._ws.close_code or 1006, msg)
                await self._handle_error(exc)
            else:
                raise ValueError(wsmsg)

    @staticmethod
    async def _handle_data(data: bytes) -> Optional[List[Dict[str, Any]]]:
        loop = asyncio.get_running_loop()
        op, msg = await loop.run_in_executor(None, Frame.decode, data)

        if op == WS.OP_MESSAGE:
            msg = cast(List[str], msg)
            return [json.loads(m) for m in msg]
        elif op == WS.OP_HEARTBEAT_REPLY:
            return None
        else:
            return None

    async def _handle_error(self, exc: Exception) -> None:
        logger.debug(f'Failed to receive message due to: {repr(exc)}')
        await self._emit('error_occurred', exc)
        await self._retry()

    async def _retry(self) -> None:
        if self._retry_count < self._MAX_RETRIES:
            if self._retry_delay > 0:
                logger.debug('Retry after {} second{}'.format(
                    self._retry_delay,
                    's' if self._retry_delay > 1 else '',
                ))
                await asyncio.sleep(self._retry_delay)
            await self.reconnect()
            self._retry_count += 1
            self._retry_delay += 1
        else:
            raise aiohttp.WebSocketError(1006, 'Over the maximum of retries')


class Frame:
    HEADER_FORMAT = '>IHHII'

    @staticmethod
    def encode(op: int, msg: str) -> bytes:
        body = msg.encode()
        header_length = WS.PACKAGE_HEADER_TOTAL_LENGTH
        packet_length = header_length + len(body)
        ver = WS.HEADER_DEFAULT_VERSION
        seq = WS.HEADER_DEFAULT_SEQUENCE

        header = struct.pack(
            Frame.HEADER_FORMAT,
            packet_length,
            header_length,
            ver,  # protocal version
            op,  # operation
            seq,  # sequence id
        )

        return header + body

    @staticmethod
    def decode(data: bytes) -> Tuple[int, Union[int, str, List[str]]]:
        plen, hlen, ver, op, _ = struct.unpack_from(
            Frame.HEADER_FORMAT, data, 0
        )
        body = data[hlen:]

        if op == WS.OP_MESSAGE:
            if ver == WS.BODY_PROTOCOL_VERSION_DEFLATE:
                data = zlib.decompress(body)

            msg_list = []
            offset = 0
            while offset < len(data):
                plen, hlen, ver, op, _ = struct.unpack_from(
                    Frame.HEADER_FORMAT, data, offset
                )
                body = data[hlen + offset:plen + offset]
                msg = body.decode('utf8')
                msg_list.append(msg)
                offset += plen

            return op, msg_list
        elif op == WS.OP_HEARTBEAT_REPLY:
            online_count = cast(int, struct.unpack('>I', body)[0])
            return op, online_count
        elif op == WS.OP_CONNECT_SUCCESS:
            auth_result = body.decode()
            return op, auth_result
        else:
            raise ValueError(f'Unexpected Operation: {op}')


class WS(IntEnum):
    AUTH_OK = 0
    AUTH_TOKEN_ERROR = -101
    BODY_PROTOCOL_VERSION_DEFLATE = 2
    BODY_PROTOCOL_VERSION_NORMAL = 0
    HEADER_DEFAULT_OPERATION = 1
    HEADER_DEFAULT_SEQUENCE = 1
    HEADER_DEFAULT_VERSION = 1
    HEADER_OFFSET = 4
    OP_CONNECT_SUCCESS = 8
    OP_HEARTBEAT = 2
    OP_HEARTBEAT_REPLY = 3
    OP_MESSAGE = 5
    OP_USER_AUTHENTICATION = 7
    OPERATION_OFFSET = 8
    PACKAGE_HEADER_TOTAL_LENGTH = 16
    PACKAGE_OFFSET = 0
    SEQUENCE_OFFSET = 12
    VERSION_OFFSET = 6


class DanmakuCommand(Enum):
    ACTIVITY_MATCH_GIFT = 'ACTIVITY_MATCH_GIFT'
    ANCHOR_LOT_AWARD = 'ANCHOR_LOT_AWARD'
    ANCHOR_LOT_CHECKSTATUS = 'ANCHOR_LOT_CHECKSTATUS'
    ANCHOR_LOT_END = 'ANCHOR_LOT_END'
    ANCHOR_LOT_START = 'ANCHOR_LOT_START'
    ANIMATION = 'ANIMATION'
    BOX_ACTIVITY_START = 'BOX_ACTIVITY_START'
    CHANGE_ROOM_INFO = 'CHANGE_ROOM_INFO'
    CHASE_FRAME_SWITCH = 'CHASE_FRAME_SWITCH'
    COMBO_SEND = 'COMBO_SEND'
    CUT_OFF = 'CUT_OFF'
    DANMU_GIFT_LOTTERY_AWARD = 'DANMU_GIFT_LOTTERY_AWARD'
    DANMU_GIFT_LOTTERY_END = 'DANMU_GIFT_LOTTERY_END'
    DANMU_GIFT_LOTTERY_START = 'DANMU_GIFT_LOTTERY_START'
    DANMU_MSG = 'DANMU_MSG'
    ENTRY_EFFECT = 'ENTRY_EFFECT'
    GUARD_ACHIEVEMENT_ROOM = 'GUARD_ACHIEVEMENT_ROOM'
    GUARD_LOTTERY_START = 'GUARD_LOTTERY_START'
    HOUR_RANK_AWARDS = 'HOUR_RANK_AWARDS'
    LITTLE_TIPS = 'LITTLE_TIPS'
    LIVE = 'LIVE'
    LOL_ACTIVITY = 'LOL_ACTIVITY'
    LUCK_GIFT_AWARD_USER = 'LUCK_GIFT_AWARD_USER'
    MATCH_TEAM_GIFT_RANK = 'MATCH_TEAM_GIFT_RANK'
    MESSAGEBOX_USER_GAIN_MEDAL = 'MESSAGEBOX_USER_GAIN_MEDAL'
    NOTICE_MSG = 'NOTICE_MSG'
    PK_AGAIN = 'PK_AGAIN'
    PK_BATTLE_CRIT = 'PK_BATTLE_CRIT'
    PK_BATTLE_END = 'PK_BATTLE_END'
    PK_BATTLE_GIFT = 'PK_BATTLE_GIFT'
    PK_BATTLE_PRE = 'PK_BATTLE_PRE'
    PK_BATTLE_PRO_TYPE = 'PK_BATTLE_PRO_TYPE'
    PK_BATTLE_PROCESS = 'PK_BATTLE_PROCESS'
    PK_BATTLE_RANK_CHANGE = 'PK_BATTLE_RANK_CHANGE'
    PK_BATTLE_SETTLE_USER = 'PK_BATTLE_SETTLE_USER'
    PK_BATTLE_SPECIAL_GIFT = 'PK_BATTLE_SPECIAL_GIFT'
    PK_BATTLE_START = 'PK_BATTLE_START'
    PK_BATTLE_VOTES_ADD = 'PK_BATTLE_VOTES_ADD'
    PK_END = 'PK_END'
    PK_LOTTERY_START = 'PK_LOTTERY_START'
    PK_MATCH = 'PK_MATCH'
    PK_MIC_END = 'PK_MIC_END'
    PK_PRE = 'PK_PRE'
    PK_PROCESS = 'PK_PROCESS'
    PK_SETTLE = 'PK_SETTLE'
    PK_START = 'PK_START'
    PREPARING = 'PREPARING'
    RAFFLE_END = 'RAFFLE_END'
    RAFFLE_START = 'RAFFLE_START'
    ROOM_BLOCK_INTO = 'ROOM_BLOCK_INTO'
    ROOM_BLOCK_MSG = 'ROOM_BLOCK_MSG'
    ROOM_BOX_USER = 'ROOM_BOX_USER'
    ROOM_CHANGE = 'ROOM_CHANGE'
    ROOM_KICKOUT = 'ROOM_KICKOUT'
    ROOM_LIMIT = 'ROOM_LIMIT'
    ROOM_LOCK = 'ROOM_LOCK'
    ROOM_RANK = 'ROOM_RANK'
    ROOM_REAL_TIME_MESSAGE_UPDATE = 'ROOM_REAL_TIME_MESSAGE_UPDATE'
    ROOM_REFRESH = 'ROOM_REFRESH'
    ROOM_SILENT_OFF = 'ROOM_SILENT_OFF'
    ROOM_SILENT_ON = 'ROOM_SILENT_ON'
    ROOM_SKIN_MSG = 'ROOM_SKIN_MSG'
    SCORE_CARD = 'SCORE_CARD'
    SEND_GIFT = 'SEND_GIFT'
    SEND_TOP = 'SEND_TOP'
    SPECIAL_GIFT = 'SPECIAL_GIFT'
    SUPER_CHAT_ENTRANCE = 'SUPER_CHAT_ENTRANCE'
    SUPER_CHAT_MESSAGE = 'SUPER_CHAT_MESSAGE'
    SUPER_CHAT_MESSAGE_DELETE = 'SUPER_CHAT_MESSAGE_DELETE'
    TV_END = 'TV_END'
    TV_START = 'TV_START'
    USER_TOAST_MSG = 'USER_TOAST_MSG'
    VOICE_JOIN_STATUS = 'VOICE_JOIN_STATUS'
    WARNING = 'WARNING'
    WATCH_LPL_EXPIRED = 'WATCH_LPL_EXPIRED'
    WEEK_STAR_CLOCK = 'WEEK_STAR_CLOCK'
    WELCOME = 'WELCOME'
    WELCOME_GUARD = 'WELCOME_GUARD'
    WIN_ACTIVITY = 'WIN_ACTIVITY'
    WIN_ACTIVITY_USER = 'WIN_ACTIVITY_USER'
    WISH_BOTTLE = 'WISH_BOTTLE'
    # ...


if __name__ == '__main__':
    def add_task_name(record: logging.LogRecord) -> bool:
        try:
            task = asyncio.current_task()
            assert task is not None
        except Exception:
            name = ''
        else:
            name = task.get_name()

        if '::' in name:
            record.roomid = '[' + name.split('::')[-1] + '] '  # type: ignore
        else:
            record.roomid = ''  # type: ignore

        return True

    def configure_logger() -> None:
        # config root logger
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        # config formatter
        fmt = '[%(asctime)s] [%(levelname)s] %(roomid)s%(message)s'
        formatter = logging.Formatter(fmt)

        # logging to console
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        console_handler.setFormatter(formatter)
        console_handler.addFilter(add_task_name)
        logger.addHandler(console_handler)

    async def get_room_list(
        session: ClientSession, count: int = 99
    ) -> List[Dict[str, Any]]:
        url = 'https://api.live.bilibili.com/room/v3/area/getRoomList'
        params = {
            'platform': 'web',
            'parent_area_id': 1,
            'cate_id': 0,
            'area_id': 0,
            'sort_type': 'online',
            'page': 1,
            'page_size': count,
            'tag_version': 1,
        }
        async with session.get(url, params=params) as response:
            j = await response.json()
            return j['data']['list']

    class DanmakuPrinter(DanmakuListener):
        def __init__(
            self, danmaku_client: DanmakuClient, room_id: int
        ) -> None:
            self._danmaku_client = danmaku_client
            self._room_id = room_id

        async def enable(self) -> None:
            self._danmaku_client.add_listener(self)

        async def disable(self) -> None:
            self._danmaku_client.remove_listener(self)

        async def on_danmaku_received(self, danmu: Danmaku) -> None:
            json_string = json.dumps(danmu, ensure_ascii=False)
            logger.info(f'{json_string}')

    class DanmakuDumper(DanmakuListener):
        def __init__(
            self, danmaku_client: DanmakuClient, room_id: int
        ) -> None:
            self._danmaku_client = danmaku_client
            from datetime import datetime
            data_time_string = datetime.now().strftime("%Y-%m-%d_%H%M%S")
            self._filename = f'blive_cmd_{room_id}_{data_time_string}.jsonl'

        async def start(self) -> None:
            import aiofiles
            self._file = await aiofiles.open(
                self._filename, mode='wt', encoding='utf8'
            )
            logger.debug(f'Opened file: {self._filename}')
            self._danmaku_client.add_listener(self)

        async def stop(self) -> None:
            self._danmaku_client.remove_listener(self)
            await self._file.close()
            logger.debug(f'Closed file: {self._filename}')

        async def on_danmaku_received(self, danmu: Danmaku) -> None:
            json_string = json.dumps(danmu, ensure_ascii=False)
            await self._file.write(json_string + '\n')

    async def test_room(session: ClientSession, room_id: int) -> None:
        client = DanmakuClient(session, room_id)
        printer = DanmakuPrinter(client, room_id)
        dumper = DanmakuDumper(client, room_id)

        await printer.enable()
        await dumper.start()
        await client.start()

        keyboard_interrupt_event = asyncio.Event()
        try:
            await keyboard_interrupt_event.wait()
        finally:
            await client.stop()
            await dumper.stop()
            await printer.disable()

    async def test() -> None:
        import sys
        try:
            # the room id must not be the short id!
            room_ids = list(map(int, sys.argv[1:]))
        except ValueError:
            print('Usage: room_id, ...')
            sys.exit(0)

        configure_logger()

        async with ClientSession() as session:
            tasks = []

            if not room_ids:
                room_list = await get_room_list(session)
                room_ids = [room['roomid'] for room in room_list]

            logger.debug(f'room count: {len(room_ids)}')

            for room_id in room_ids:
                task = asyncio.create_task(
                    test_room(session, room_id), name=f'test_room::{room_id}',
                )
                tasks.append(task)

            await asyncio.wait(tasks)

    try:
        asyncio.run(test())
    except KeyboardInterrupt:
        pass



import io
import os
import time
import errno
import asyncio
import logging
from threading import Thread
from datetime import datetime, timezone, timedelta
from collections import OrderedDict

from typing import Any, BinaryIO, Dict, Iterator, Optional, Tuple

import aiohttp
import requests
import urllib3
from tqdm import tqdm
from rx.subject import Subject
from rx.core import Observable
from tenacity import (
    retry,
    wait_none,
    wait_fixed,
    wait_chain,
    wait_exponential,
    stop_after_delay,
    stop_after_attempt,
    retry_if_result,
    retry_if_exception,
    retry_if_exception_type,
    Retrying,
    TryAgain,
)

from .. import __version__, __prog__, __github__
from .retry import wait_exponential_for_same_exceptions, before_sleep_log
from .statistics import StatisticsCalculator
from ..event.event_emitter import EventListener, EventEmitter
from ..bili.live import Live
from ..bili.typing import QualityNumber
from ..flv.stream_processor import StreamProcessor, BaseOutputFileManager
from ..utils.mixins import AsyncCooperationMix, AsyncStoppableMixin
from ..flv.exceptions import FlvStreamCorruptedError
from ..bili.exceptions import (
    LiveRoomHidden, LiveRoomLocked, LiveRoomEncrypted, NoStreamUrlAvailable
)


__all__ = 'StreamRecorderEventListener', 'StreamRecorder'


logger = logging.getLogger(__name__)
logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


class StreamRecorderEventListener(EventListener):
    async def on_video_file_created(
        self, path: str, record_start_time: int
    ) -> None:
        ...

    async def on_video_file_completed(self, path: str) -> None:
        ...


class StreamRecorder(
    EventEmitter[StreamRecorderEventListener],
    AsyncCooperationMix,
    AsyncStoppableMixin,
):
    def __init__(
        self,
        live: Live,
        out_dir: str,
        path_template: str,
        *,
        quality_number: QualityNumber = 10000,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
    ) -> None:
        super().__init__()

        self._live = live
        self._progress_bar: Optional[tqdm] = None
        self._stream_processor: Optional[StreamProcessor] = None
        self._calculator = StatisticsCalculator()
        self._file_manager = OutputFileManager(
            live, out_dir, path_template, buffer_size
        )

        self._quality_number = quality_number
        self._real_quality_number: Optional[QualityNumber] = None
        self.buffer_size = buffer_size or io.DEFAULT_BUFFER_SIZE  # bytes
        self.read_timeout = read_timeout or 3  # seconds

        self._filesize_limit = filesize_limit or 0
        self._duration_limit = duration_limit or 0

        def on_file_created(args: Tuple[str, int]) -> None:
            logger.info(f"Video file created: '{args[0]}'")
            self._emit_event('video_file_created', *args)

        def on_file_closed(path: str) -> None:
            logger.info(f"Video file completed: '{path}'")
            self._emit_event('video_file_completed', path)

        self._file_manager.file_creates.subscribe(on_file_created)
        self._file_manager.file_closes.subscribe(on_file_closed)

    @property
    def data_count(self) -> int:
        return self._calculator.count

    @property
    def data_rate(self) -> float:
        return self._calculator.rate

    @property
    def elapsed(self) -> float:
        return self._calculator.elapsed

    @property
    def out_dir(self) -> str:
        return self._file_manager.out_dir

    @out_dir.setter
    def out_dir(self, value: str) -> None:
        self._file_manager.out_dir = value

    @property
    def path_template(self) -> str:
        return self._file_manager.path_template

    @path_template.setter
    def path_template(self, value: str) -> None:
        self._file_manager.path_template = value

    @property
    def quality_number(self) -> QualityNumber:
        return self._quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._quality_number = value
        self._real_quality_number = None

    @property
    def real_quality_number(self) -> QualityNumber:
        return self._real_quality_number or 10000

    @property
    def filesize_limit(self) -> int:
        if self._stream_processor is not None:
            return self._stream_processor.filesize_limit
        else:
            return self._filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._filesize_limit = value
        if self._stream_processor is not None:
            self._stream_processor.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        if self._stream_processor is not None:
            return self._stream_processor.duration_limit
        else:
            return self._duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._duration_limit = value
        if self._stream_processor is not None:
            self._stream_processor.duration_limit = value

    def has_file(self) -> bool:
        return self._file_manager.has_file()

    def get_files(self) -> Iterator[str]:
        yield from self._file_manager.get_files()

    def clear_files(self) -> None:
        self._file_manager.clear_files()

    def update_progress_bar_info(self) -> None:
        if self._progress_bar is not None:
            self._progress_bar.set_postfix_str(self._make_pbar_postfix())

    async def _do_start(self) -> None:
        self._thread = Thread(
            target=self._run, name=f'StreamRecorder::{self._live.room_id}'
        )
        self._thread.start()
        logger.debug('Started stream recorder')

    async def _do_stop(self) -> None:
        logger.debug('Stopping stream recorder...')
        if self._stream_processor is not None:
            self._stream_processor.cancel()
        await self._loop.run_in_executor(None, self._thread.join)
        logger.debug('Stopped stream recorder')

    def _run(self) -> None:
        self._calculator.reset()
        try:
            with tqdm(
                desc='Recording',
                unit='B',
                unit_scale=True,
                unit_divisor=1024,
                postfix=self._make_pbar_postfix(),
            ) as progress_bar:
                self._progress_bar = progress_bar

                def update_size(size: int) -> None:
                    progress_bar.update(size)
                    self._calculator.submit(size)

                self._stream_processor = StreamProcessor(
                    self._file_manager,
                    filesize_limit=self._filesize_limit,
                    duration_limit=self._duration_limit,
                    metadata=self._make_metadata(),
                    analyse_data=True,
                    dedup_join=True,
                    save_extra_metadata=True,
                )
                self._stream_processor.size_updates.subscribe(update_size)

                with requests.Session() as self._session:
                    self._main_loop()
        except TryAgain:
            pass
        except Exception as e:
            self._handle_exception(e)
        finally:
            if self._stream_processor is not None:
                self._stream_processor.finalize()
                self._stream_processor = None
            self._progress_bar = None
            self._calculator.freeze()

    def _main_loop(self) -> None:
        for attempt in Retrying(
            reraise=True,
            retry=(
                retry_if_result(lambda r: not self._stopped) |
                retry_if_exception(lambda e: not isinstance(e, OSError))
            ),
            wait=wait_exponential_for_same_exceptions(max=60),
            before_sleep=before_sleep_log(logger, logging.DEBUG, 'main_loop'),
        ):
            with attempt:
                try:
                    self._streaming_loop()
                except NoStreamUrlAvailable:
                    logger.debug('No stream url available')
                    if not self._stopped:
                        raise TryAgain
                except OSError as e:
                    if e.errno == errno.ENOSPC:
                        # OSError(28, 'No space left on device')
                        raise
                    logger.critical(repr(e), exc_info=e)
                    raise TryAgain
                except LiveRoomHidden:
                    logger.error('The live room has been hidden!')
                    self._stopped = True
                except LiveRoomLocked:
                    logger.error('The live room has been locked!')
                    self._stopped = True
                except LiveRoomEncrypted:
                    logger.error('The live room has been encrypted!')
                    self._stopped = True
                except Exception as e:
                    logger.exception(e)
                    raise

    def _streaming_loop(self) -> None:
        url = self._get_live_stream_url()

        while not self._stopped:
            try:
                self._streaming(url)
            except requests.exceptions.HTTPError as e:
                # frequently occurred when the live just started or ended.
                logger.debug(repr(e))
                self._defer_retry(1, 'streaming_loop')
                # the url may has been forbidden or expired
                # when the status code is 404 or 403
                if e.response.status_code in (403, 404):
                    url = self._get_live_stream_url()
            except requests.exceptions.Timeout as e:
                logger.warning(repr(e))
            except urllib3.exceptions.TimeoutError as e:
                logger.warning(repr(e))
            except urllib3.exceptions.ProtocolError as e:
                # ProtocolError('Connection broken: IncompleteRead(
                logger.warning(repr(e))
            except requests.exceptions.ConnectionError as e:
                logger.warning(repr(e))
            except FlvStreamCorruptedError as e:
                logger.warning(repr(e))

    def _streaming(self, url: str) -> None:
        logger.debug('Getting the live stream...')
        with self._session.get(
            url,
            headers=self._live.headers,
            stream=True,
            timeout=self.read_timeout,
        ) as response:
            logger.debug('Response received')
            response.raise_for_status()

            assert self._stream_processor is not None
            self._stream_processor.process_stream(
                io.BufferedReader(
                    ResponseProxy(response.raw), buffer_size=8192
                )
            )

    @retry(
        reraise=True,
        retry=retry_if_exception_type((
            asyncio.TimeoutError, aiohttp.ClientError,
        )),
        wait=wait_chain(wait_none(), wait_fixed(1)),
        stop=stop_after_attempt(300),
    )
    def _get_live_stream_url(self) -> str:
        qn = self._real_quality_number or self.quality_number
        url = self._run_coroutine(self._live.get_live_stream_url(qn, 'flv'))

        if self._real_quality_number is None:
            if url is None:
                logger.info(
                    f'The specified video quality ({qn}) is not available, '
                    'using the original video quality (10000) instead.'
                )
                self._real_quality_number = 10000
                raise TryAgain
            else:
                logger.info(f'The specified video quality ({qn}) is available')
                self._real_quality_number = self.quality_number

        assert url is not None
        logger.debug(f"Got live stream url: '{url}'")

        return url

    def _defer_retry(self, seconds: float, name: str = '') -> None:
        if seconds <= 0:
            return
        logger.debug(f'Retry {name} after {seconds} seconds')
        time.sleep(seconds)

    def _make_pbar_postfix(self) -> str:
        return '{room_id} - {user_name}: {room_title}'.format(
            room_id=self._live.room_info.room_id,
            user_name=self._live.user_info.name,
            room_title=self._live.room_info.title,
        )

    def _make_metadata(self) -> Dict[str, Any]:
        live_start_time = datetime.fromtimestamp(
            self._live.room_info.live_start_time, timezone(timedelta(hours=8))
        )

        return {
            'Title': self._live.room_info.title,
            'Artist': self._live.user_info.name,
            'Date': str(live_start_time),
            'Comment': f'''\
B站直播录像
主播：{self._live.user_info.name}
标题：{self._live.room_info.title}
分区：{self._live.room_info.parent_area_name} - {self._live.room_info.area_name}
房间号：{self._live.room_info.room_id}
开播时间：{live_start_time}
录制程序：{__prog__} v{__version__} {__github__}''',
            'description': OrderedDict({
                'UserId': str(self._live.user_info.uid),
                'UserName': self._live.user_info.name,
                'RoomId': str(self._live.room_info.room_id),
                'RoomTitle': self._live.room_info.title,
                'Area': self._live.room_info.area_name,
                'ParentArea': self._live.room_info.parent_area_name,
                'LiveStartTime': str(live_start_time),
                'Recorder': f'{__prog__} v{__version__} {__github__}',
            })
        }

    def _emit_event(self, name: str, *args: Any, **kwds: Any) -> None:
        self._run_coroutine(super()._emit(name, *args, **kwds))


class ResponseProxy(io.RawIOBase):
    def __init__(self, response: urllib3.HTTPResponse) -> None:
        self._response = response

    @property
    def closed(self) -> bool:
        # always return False to avoid that `ValueError: read of closed file`,
        # raised from `CHECK_CLOSED(self, "read of closed file")`,
        # result in losing data those remaining in the buffer.
        # ref: `https://gihub.com/python/cpython/blob/63298930fb531ba2bb4f23bc3b915dbf1e17e9e1/Modules/_io/bufferedio.c#L882`  # noqa
        return False

    def readable(self) -> bool:
        return True

    def read(self, size: int = -1) -> bytes:
        return self._response.read(size)

    def tell(self) -> int:
        return self._response.tell()

    def readinto(self, b: Any) -> int:
        return self._response.readinto(b)

    def close(self) -> None:
        self._response.close()


class OutputFileManager(BaseOutputFileManager, AsyncCooperationMix):
    def __init__(
        self,
        live: Live,
        out_dir: str,
        path_template: str,
        buffer_size: Optional[int] = None,
    ) -> None:
        super().__init__(buffer_size)
        self._live = live

        self.out_dir = out_dir
        self.path_template = path_template

        self._file_creates = Subject()
        self._file_closes = Subject()

    @property
    def file_creates(self) -> Observable:
        return self._file_creates

    @property
    def file_closes(self) -> Observable:
        return self._file_closes

    def create_file(self) -> BinaryIO:
        self._start_time = self._get_timestamp()
        file = super().create_file()
        self._file_creates.on_next((self._curr_path, self._start_time))
        return file

    def close_file(self) -> None:
        path = self._curr_path
        super().close_file()
        self._file_closes.on_next(path)

    def _get_timestamp(self) -> int:
        try:
            return self._get_server_timestamp()
        except Exception as e:
            logger.warning(f'Failed to get server timestamp: {repr(e)}')
            return self._get_local_timestamp()

    def _get_local_timestamp(self) -> int:
        return int(time.time())

    @retry(
        reraise=True,
        retry=retry_if_exception_type((
            asyncio.TimeoutError, aiohttp.ClientError,
        )),
        wait=wait_exponential(multiplier=0.1, max=1),
        stop=stop_after_delay(3),
    )
    def _get_server_timestamp(self) -> int:
        return self._run_coroutine(self._live.get_server_timestamp())

    def _make_path(self) -> str:
        date_time = datetime.fromtimestamp(self._start_time)
        relpath = self.path_template.format(
            roomid=self._live.room_id,
            uname=self._live.user_info.name,
            title=self._live.room_info.title,
            area=self._live.room_info.area_name,
            parent_area=self._live.room_info.parent_area_name,
            year=date_time.year,
            month=str(date_time.month).rjust(2, '0'),
            day=str(date_time.day).rjust(2, '0'),
            hour=str(date_time.hour).rjust(2, '0'),
            minute=str(date_time.minute).rjust(2, '0'),
            second=str(date_time.second).rjust(2, '0'),
        )

        full_pathname = os.path.abspath(
            os.path.expanduser(os.path.join(self.out_dir, relpath) + '.flv')
        )
        os.makedirs(os.path.dirname(full_pathname), exist_ok=True)

        return full_pathname



from __future__ import annotations
import logging
from typing import Optional


from .models import AudioTag, FlvTag, ScriptTag, VideoTag
from .common import (
    is_audio_sequence_header, is_video_sequence_header, is_metadata_tag
)
from .exceptions import AudioParametersChanged, VideoParametersChanged


__all__ = 'ParametersChecker',


logger = logging.getLogger(__name__)


class ParametersChecker:
    def __init__(self) -> None:
        self.reset()

    @property
    def last_metadata_tag(self) -> Optional[ScriptTag]:
        return self._last_metadata_tag

    @property
    def last_audio_header_tag(self) -> Optional[AudioTag]:
        return self._last_audio_header_tag

    @property
    def last_video_header_tag(self) -> Optional[VideoTag]:
        return self._last_video_header_tag

    def reset(self) -> None:
        self._last_metadata_tag: Optional[ScriptTag] = None
        self._last_audio_header_tag: Optional[AudioTag] = None
        self._last_video_header_tag: Optional[VideoTag] = None

    def check_tag(self, tag: FlvTag) -> None:
        if is_audio_sequence_header(tag):
            if self._last_audio_header_tag is not None:
                if not tag.is_the_same_as(self._last_audio_header_tag):
                    logger.warning('Audio parameters changed')
                    self._last_audio_header_tag = tag
                    raise AudioParametersChanged()
            self._last_audio_header_tag = tag
        elif is_video_sequence_header(tag):
            if self._last_video_header_tag is not None:
                if not tag.is_the_same_as(self._last_video_header_tag):
                    logger.warning('Video parameters changed')
                    self._last_video_header_tag = tag
                    raise VideoParametersChanged()
            self._last_video_header_tag = tag
        elif is_metadata_tag(tag):
            self._last_metadata_tag = tag
        else:
            pass



from __future__ import annotations
import html
import asyncio
from typing import AsyncIterator, Final, List

from lxml import etree
import aiofiles

from .models import Metadata, Danmu
from .typing import Element


__all__ = 'DanmakuReader', 'DanmakuWriter'


class DanmakuReader:
    def __init__(self, path: str) -> None:
        self._path = path

    async def __aenter__(self) -> DanmakuReader:
        await self.init()
        return self

    async def __aexit__(self, exc_type, exc, tb):  # type: ignore
        pass

    async def init(self) -> None:
        loop = asyncio.get_running_loop()
        self._tree = await loop.run_in_executor(None, etree.parse, self._path)

    async def read_metadata(self) -> Metadata:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._read_metadata)

    async def read_danmus(self) -> AsyncIterator[Danmu]:
        for elem in await self._get_danmu_elems():
            yield self._make_danmu(elem)

    def _read_metadata(self) -> Metadata:
        return Metadata(
            user_name=self._tree.xpath('/i/metadata/user_name')[0].text,
            room_id=int(self._tree.xpath('/i/metadata/room_id')[0].text),
            room_title=self._tree.xpath('/i/metadata/room_title')[0].text,
            area=self._tree.xpath('/i/metadata/area')[0].text,
            parent_area=self._tree.xpath('/i/metadata/parent_area')[0].text,
            live_start_time=int(
                self._tree.xpath('/i/metadata/live_start_time')[0].text
            ),
            record_start_time=int(
                self._tree.xpath('/i/metadata/record_start_time')[0].text
            ),
            recorder=self._tree.xpath('/i/metadata/recorder')[0].text,
        )

    async def _get_danmu_elems(self) -> List[Element]:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._tree.xpath, '/i/d')

    def _make_danmu(self, elem: Element) -> Danmu:
        params = elem.get('p').split(',')
        return Danmu(
            stime=float(params[0]),
            mode=int(params[1]),
            size=int(params[2]),
            color=int(params[3]),
            date=int(params[4]),
            pool=int(params[5]),
            uid=params[6],
            dmid=int(params[7]),
            text=elem.text,
        )


class DanmakuWriter:
    _XML_HEAD: Final[str] = """\
<?xml version="1.0" encoding="UTF-8"?>
<i>
    <chatserver>chat.bilibili.com</chatserver>
    <chatid>0</chatid>
    <mission>0</mission>
    <maxlimit>0</maxlimit>
    <state>0</state>
    <real_name>0</real_name>
    <source>e-r</source>
"""

    def __init__(self, path: str):
        self._path = path

    async def __aenter__(self) -> DanmakuWriter:
        await self.init()
        return self

    async def __aexit__(self, exc_type, exc, tb):  # type: ignore
        await self.complete()

    async def init(self) -> None:
        self._file = await aiofiles.open(self._path, 'wt', encoding='utf8')
        await self._file.write(self._XML_HEAD)

    async def write_metadata(self, metadata: Metadata) -> None:
        await self._file.write(self._serialize_metadata(metadata))

    async def write_danmu(self, danmu: Danmu) -> None:
        await self._file.write(self._serialize_danmu(danmu))

    async def complete(self) -> None:
        await self._file.write('</i>')
        await self._file.close()

    def _serialize_metadata(self, metadata: Metadata) -> str:
        return f"""\
    <metadata>
        <user_name>{html.escape(metadata.user_name)}</user_name>
        <room_id>{metadata.room_id}</room_id>
        <room_title>{html.escape(metadata.room_title)}</room_title>
        <area>{html.escape(metadata.area)}</area>
        <parent_area>{html.escape(metadata.parent_area)}</parent_area>
        <live_start_time>{metadata.live_start_time}</live_start_time>
        <record_start_time>{metadata.record_start_time}</record_start_time>
        <recorder>{metadata.recorder}</recorder>
    </metadata>
"""

    def _serialize_danmu(self, dm: Danmu) -> str:
        return (
            f'    <d p="{dm.stime:.5f},{dm.mode},{dm.size},{dm.color},'
            f'{dm.date},{dm.pool},{dm.uid},{dm.dmid}">'
            f'{html.escape(dm.text)}</d>\n'
        )



from __future__ import annotations
import os
from enum import Enum
from typing import Optional

import attr

from ..bili.models import RoomInfo, UserInfo
from ..bili.typing import QualityNumber
from ..postprocess import DeleteStrategy
from ..postprocess.typing import Progress


class RunningStatus(str, Enum):
    STOPPED = 'stopped'
    WAITING = 'waiting'
    RECORDING = 'recording'
    REMUXING = 'remuxing'
    INJECTING = 'injecting'


@attr.s(auto_attribs=True, slots=True, frozen=True)
class TaskStatus:
    monitor_enabled: bool
    recorder_enabled: bool
    running_status: RunningStatus
    elapsed: float  # time elapsed
    data_count: int  # Number of Bytes in total
    data_rate: float  # Number of Bytes per second
    danmu_count: int  # Number of Danmu in total
    danmu_rate: float  # Number of Danmu per minutes
    real_quality_number: QualityNumber
    postprocessing_path: Optional[str] = None
    postprocessing_progress: Optional[Progress] = None


@attr.s(auto_attribs=True, slots=True, frozen=True)
class TaskParam:
    # OutputSettings
    out_dir: str
    path_template: str
    filesize_limit: int
    duration_limit: int
    # HeaderSettings
    user_agent: str
    cookie: str
    # DanmakuSettings
    danmu_uname: bool
    # RecorderSettings
    quality_number: QualityNumber
    read_timeout: int
    buffer_size: int
    # PostprocessingOptions
    remux_to_mp4: bool
    delete_source: DeleteStrategy


@attr.s(auto_attribs=True, slots=True, frozen=True)
class TaskData:
    user_info: UserInfo
    room_info: RoomInfo
    task_status: TaskStatus


@attr.s(auto_attribs=True, slots=True, frozen=True)
class FileDetail:
    exists: bool
    path: str
    size: int

    @classmethod
    def from_path(cls, path: str) -> FileDetail:
        try:
            return cls(True, path, os.path.getsize(path))
        except FileNotFoundError:
            return cls(False, path, 0)



from typing import Union


from .models import (
    LiveBeganEvent,
    LiveBeganEventData,
    LiveEndedEvent,
    LiveEndedEventData,
    RoomChangeEvent,
    RoomChangeEventData,
    SpaceNoEnoughEvent,
    SpaceNoEnoughEventData,
    FileCompletedEvent,
    FileCompletedEventData,
    Error,
    ErrorData,
)


Event = Union[
    LiveBeganEvent,
    LiveEndedEvent,
    RoomChangeEvent,
    SpaceNoEnoughEvent,
    FileCompletedEvent,
    Error,
]

EventData = Union[
    LiveBeganEventData,
    LiveEndedEventData,
    RoomChangeEventData,
    SpaceNoEnoughEventData,
    FileCompletedEventData,
    ErrorData,
]



import os
import logging
import secrets
from typing import Optional, Tuple

from fastapi import FastAPI, status, Request, Depends, Header
from fastapi.responses import JSONResponse
from fastapi.exceptions import HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from pydantic import ValidationError
from pkg_resources import resource_filename

from .routers import (
    tasks, settings, application, validation, websockets, update
)
from .schemas import ResponseMessage
from ..setting import EnvSettings, Settings
from ..application import Application
from ..exception import NotFoundError, ExistsError


logger = logging.getLogger(__name__)


_env_settings = EnvSettings()
_settings = Settings.load(_env_settings.settings_file)
_settings.update_from_env_settings(_env_settings)
app = Application(_settings)

if _env_settings.api_key is None:
    _dependencies = None
else:
    async def validate_api_key(
        x_api_key: Optional[str] = Header(None)
    ) -> None:
        assert _env_settings.api_key is not None
        if (
            x_api_key is None or
            not secrets.compare_digest(x_api_key, _env_settings.api_key)
        ):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='API key is missing or invalid',
            )

    _dependencies = [Depends(validate_api_key)]

api = FastAPI(
    title='Bilibili live streaming recorder web API',
    description='Web API to communicate with the backend application',
    version='v1',
    dependencies=_dependencies,
)

api.add_middleware(
    CORSMiddleware,
    allow_origins=[
        'http://localhost:4200',  # angular development
    ],
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)


@api.exception_handler(NotFoundError)
async def not_found_error_handler(
    request: Request, exc: NotFoundError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_404_NOT_FOUND,
        content=dict(ResponseMessage(
            code=status.HTTP_404_NOT_FOUND,
            message=str(exc),
        )),
    )


@api.exception_handler(ExistsError)
async def exists_error_handler(
    request: Request, exc: ExistsError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_409_CONFLICT,
        content=dict(ResponseMessage(
            code=status.HTTP_409_CONFLICT,
            message=str(exc),
        )),
    )


@api.exception_handler(ValidationError)
async def validation_error_handler(
    request: Request, exc: ValidationError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_406_NOT_ACCEPTABLE,
        content=dict(ResponseMessage(
            code=status.HTTP_406_NOT_ACCEPTABLE,
            message=str(exc),
        )),
    )


@api.on_event('startup')
async def on_startup() -> None:
    await app.launch()


@api.on_event('shutdown')
async def on_shuntdown() -> None:
    _settings.dump()
    await app.exit()


tasks.app = app
settings.app = app
application.app = app
validation.app = app
websockets.app = app
update.app = app
api.include_router(tasks.router)
api.include_router(settings.router)
api.include_router(application.router)
api.include_router(validation.router)
api.include_router(websockets.router)
api.include_router(update.router)


class WebAppFiles(StaticFiles):
    async def lookup_path(
        self, path: str
    ) -> Tuple[str, Optional[os.stat_result]]:
        if path == '404.html':
            path = 'index.html'
        return await super().lookup_path(path)


directory = resource_filename(__name__, '../data/webapp')
api.mount('/', WebAppFiles(directory=directory, html=True), name='webapp')



from __future__ import annotations
import logging
from typing import Optional


from .models import FlvHeader, FlvTag, BACK_POINTER_SIZE, VideoTag
from .exceptions import FileSizeOverLimit, DurationOverLimit
from .common import is_video_nalu_keyframe


__all__ = 'LimitChecker',


logger = logging.getLogger(__name__)


class LimitChecker:
    def __init__(
        self,
        filesize_limit: int = 0,  # file size in bytes, no limit by default.
        duration_limit: int = 0,  # duration in seconds, no limit by default.
    ) -> None:
        self.filesize_limit = filesize_limit
        self.duration_limit = duration_limit
        self.reset()

    def is_filesize_over_limit(self) -> bool:
        return (
            self._filesize + self._max_size_between_keyframes >=
            self.filesize_limit
        )

    def is_duration_over_limit(self) -> bool:
        return (
            self._duration + self._max_duration_between_keyframes >=
            self.duration_limit
        )

    @property
    def last_keyframe_tag(self) -> Optional[VideoTag]:
        return self._last_keyframe_tag

    def reset(self) -> None:
        self._filesize = 0
        self._duration = 0.0
        self._max_size_between_keyframes = 0
        self._max_duration_between_keyframes = 0.0
        self._header_checked = False
        self._last_keyframe_tag: Optional[VideoTag] = None

    def check_header(self, header: FlvHeader) -> None:
        assert not self._header_checked
        self._header_checked = True
        self._filesize += header.size + BACK_POINTER_SIZE

    def check_tag(self, tag: FlvTag) -> None:
        self._filesize += tag.tag_size + BACK_POINTER_SIZE
        self._duration = tag.timestamp / 1000

        if not is_video_nalu_keyframe(tag):
            return

        if self._last_keyframe_tag is not None:
            self._max_size_between_keyframes = max(
                self._max_size_between_keyframes,
                tag.offset - self._last_keyframe_tag.offset,
            )
            self._max_duration_between_keyframes = max(
                self._max_duration_between_keyframes,
                (tag.timestamp - self._last_keyframe_tag.timestamp) / 1000,
            )
        self._last_keyframe_tag = tag

        if self.filesize_limit > 0 and self.is_filesize_over_limit():
            logger.debug('File size will be over the limit: {} + {}'.format(
                self._filesize, self._max_size_between_keyframes,
            ))
            raise FileSizeOverLimit()

        if self.duration_limit > 0 and self.is_duration_over_limit():
            logger.debug('Duration will be over the limit: {} + {}'.format(
                self._duration, self._max_duration_between_keyframes,
            ))
            raise DurationOverLimit()



import logging
from typing import Any, Callable, Optional, Type, cast

from tenacity import wait_exponential, RetryCallState
from tenacity import _utils
from tenacity import compat as _compat


class wait_exponential_for_same_exceptions(wait_exponential):
    """Wait strategy that applies exponential backoff only for same
    continuing exceptions.
    """

    def __init__(
        self,
        multiplier: float = 1,
        max: float = _utils.MAX_WAIT,
        exp_base: int = 2,
        min: float = 0,
        continuing_criteria: float = 5.0,
    ) -> None:
        super().__init__(multiplier, max, exp_base, min)
        self._continuing_criteria = continuing_criteria
        self._prev_exc_type: Optional[Type[BaseException]] = None
        self._prev_exc_ts: Optional[float] = None
        self._last_wait_time: float = 0

    @_compat.wait_dunder_call_accept_old_params
    def __call__(self, retry_state: RetryCallState) -> float:
        if (
            retry_state.outcome is not None and
            (exc := retry_state.outcome.exception())
        ):
            curr_exc_type = type(exc)
            curr_exc_ts = cast(float, retry_state.outcome_timestamp)
            if (
                curr_exc_type is not self._prev_exc_type or
                not self._is_continuing(curr_exc_ts)
            ):
                retry_state.attempt_number = 1
                self._prev_exc_type = curr_exc_type
                self._prev_exc_ts = curr_exc_ts

        self._last_wait_time = wait_time = super().__call__(retry_state)
        return wait_time

    def _is_continuing(self, curr_exc_ts: float) -> bool:
        assert self._prev_exc_ts is not None
        return (
            curr_exc_ts - (self._prev_exc_ts + self._last_wait_time) <
            self._continuing_criteria
        )


def before_sleep_log(
    logger: logging.Logger, log_level: int, name: str = ''
) -> Callable[[RetryCallState], Any]:
    def log_it(retry_state: RetryCallState) -> None:
        seconds = cast(float, getattr(retry_state.next_action, 'sleep'))
        logger.log(log_level, 'Retry %s after %s seconds', name, seconds)

    return log_it



from abc import ABC, abstractmethod
import asyncio
from typing import Awaitable, TypeVar, final


class SwitchableMixin(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._enabled = False

    @property
    def enabled(self) -> bool:
        return self._enabled

    @final
    def enable(self) -> None:
        if self._enabled:
            return
        self._enabled = True
        self._do_enable()

    @final
    def disable(self) -> None:
        if not self._enabled:
            return
        self._enabled = False
        self._do_disable()

    @abstractmethod
    def _do_enable(self) -> None:
        ...

    @abstractmethod
    def _do_disable(self) -> None:
        ...


class StoppableMixin(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._stopped = True

    @property
    def stopped(self) -> bool:
        return self._stopped

    @final
    def start(self) -> None:
        if not self._stopped:
            return
        self._stopped = False
        self._do_start()

    @final
    def stop(self) -> None:
        if self._stopped:
            return
        self._stopped = True
        self._do_stop()

    @abstractmethod
    def _do_start(self) -> None:
        ...

    @abstractmethod
    def _do_stop(self) -> None:
        ...


class AsyncStoppableMixin(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._stopped = True

    @property
    def stopped(self) -> bool:
        return self._stopped

    @final
    async def start(self) -> None:
        if not self._stopped:
            return
        self._stopped = False
        await self._do_start()

    @final
    async def stop(self) -> None:
        if self._stopped:
            return
        self._stopped = True
        await self._do_stop()

    @abstractmethod
    async def _do_start(self) -> None:
        ...

    @abstractmethod
    async def _do_stop(self) -> None:
        ...


_T = TypeVar('_T')


class AsyncCooperationMix(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._loop = asyncio.get_running_loop()

    def _handle_exception(self, exc: BaseException) -> None:
        from ..exception import submit_exception  # XXX circular import

        async def wrapper() -> None:
            submit_exception(exc)
        self._run_coroutine(wrapper())

    def _run_coroutine(self, coro: Awaitable[_T]) -> _T:
        future = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return future.result()



from typing import Any, Final

import aiohttp
from tenacity import (
    retry,
    wait_exponential,
    stop_after_delay,
)

from .typing import QualityNumber, JsonResponse, ResponseData
from .exceptions import ApiRequestError


__all__ = 'WebApi',


class WebApi:
    BASE_API_URL: Final[str] = 'https://api.bilibili.com'
    BASE_LIVE_API_URL: Final[str] = 'https://api.live.bilibili.com'

    GET_USER_INFO_URL: Final[str] = BASE_API_URL + '/x/space/acc/info'

    ROOM_INIT_URL: Final[str] = BASE_LIVE_API_URL + '/room/v1/Room/room_init'
    GET_INFO_URL: Final[str] = BASE_LIVE_API_URL + '/room/v1/Room/get_info'
    GET_INFO_BY_ROOM_URL: Final[str] = BASE_LIVE_API_URL + \
        '/xlive/web-room/v1/index/getInfoByRoom'
    GET_ROOM_PLAY_INFO_URL: Final[str] = BASE_LIVE_API_URL + \
        '/xlive/web-room/v2/index/getRoomPlayInfo'
    GET_TIMESTAMP_URL: Final[str] = BASE_LIVE_API_URL + \
        '/av/v1/Time/getTimestamp?platform=pc'

    def __init__(self, session: aiohttp.ClientSession):
        self._session = session
        self.timeout = 10

    @staticmethod
    def _check_response(json_res: JsonResponse) -> None:
        if json_res['code'] != 0:
            raise ApiRequestError(
                json_res['code'],
                json_res.get('message') or json_res.get('msg') or '',
            )

    @retry(
        reraise=True,
        stop=stop_after_delay(5),
        wait=wait_exponential(0.1),
    )
    async def _get(self, *args: Any, **kwds: Any) -> JsonResponse:
        async with self._session.get(
            *args,
            **kwds,
            timeout=self.timeout,
        ) as res:
            json_res = await res.json()
            self._check_response(json_res)
            return json_res

    async def room_init(self, room_id: int) -> ResponseData:
        r = await self._get(self.ROOM_INIT_URL, params={'id': room_id})
        return r['data']

    async def get_room_play_info(
        self, room_id: int, qn: QualityNumber = 10000
    ) -> ResponseData:
        params = {
            'room_id': room_id,
            'protocol': '0,1',
            'format': '0,2',
            'codec': '0,1',
            'qn': qn,
            'platform': 'web',
            'ptype': 16,
        }
        r = await self._get(self.GET_ROOM_PLAY_INFO_URL, params=params)
        return r['data']

    async def get_info_by_room(self, room_id: int) -> ResponseData:
        params = {
            'room_id': room_id,
        }
        r = await self._get(self.GET_INFO_BY_ROOM_URL, params=params)
        return r['data']

    async def get_info(self, room_id: int) -> ResponseData:
        params = {
            'room_id': room_id,
        }
        r = await self._get(self.GET_INFO_URL, params=params)
        return r['data']

    async def get_timestamp(self) -> int:
        r = await self._get(self.GET_TIMESTAMP_URL)
        return r['data']['timestamp']

    async def get_user_info(self, uid: int) -> ResponseData:
        params = {
            'mid': uid,
        }
        r = await self._get(self.GET_USER_INFO_URL, params=params)
        return r['data']





class FlvStreamCorruptedError(Exception):
    ...


class FlvFileCorruptedError(Exception):
    ...


class AudioParametersChanged(Exception):
    ...


class VideoParametersChanged(Exception):
    ...


class FileSizeOverLimit(Exception):
    ...


class DurationOverLimit(Exception):
    ...
